from office365.runtime.client_object import ClientObject
from office365.runtime.client_object_collection import ClientObjectCollection
from office365.runtime.client_request import ClientRequest
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value import ClientValue
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.http.request_options import RequestOptions
from office365.runtime.odata.v3.json_light_format import JsonLightFormat
from office365.runtime.queries.create_entity_query import CreateEntityQuery
from office365.runtime.queries.delete_entity_query import DeleteEntityQuery
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.queries.update_entity_query import UpdateEntityQuery


class ODataRequest(ClientRequest):

    def __init__(self, context, json_format):
        """

        :type context: office365.runtime.client_runtime_context.ClientRuntimeContext
        :type json_format: office365.runtime.odata.odata_json_format.ODataJsonFormat
        """
        super(ODataRequest, self).__init__(context)
        self._json_format = json_format

    @property
    def json_format(self):
        return self._json_format

    def execute_request_direct(self, request,**kwargs):
        """

        :type request: office365.runtime.http.request_options.RequestOptions
        """
        self.ensure_media_type(request)
        return super(ODataRequest, self).execute_request_direct(request,**kwargs)

    def build_request(self):
        query = self.context.current_query
        action_url = query.build_url()
        request = RequestOptions(action_url)
        self.ensure_media_type(request)
        # set method
        request.method = HttpMethod.Get
        if isinstance(query, DeleteEntityQuery):
            request.method = HttpMethod.Post
        elif isinstance(query, (CreateEntityQuery, UpdateEntityQuery, ServiceOperationQuery)):
            request.method = HttpMethod.Post
            if query.parameter_type is not None:
                request.data = self._normalize_payload(query)
        return request

    def ensure_media_type(self, request):
        """
        :type request: RequestOptions
        """
        media_type = self.json_format.get_media_type()
        request.ensure_header('Content-Type', media_type)
        request.ensure_header('Accept', media_type)

    def process_response(self, response):
        """
        :type response: requests.Response
        """
        query = self.context.current_query
        return_type = query.return_type
        if return_type is None:
            return

        if isinstance(return_type, ClientObject):
            return_type.clear()

        if response.headers.get('Content-Type', '').lower().split(';')[0] != 'application/json':
            if isinstance(return_type, ClientResult):
                return_type.value = response.content
        else:
            if isinstance(query, ServiceOperationQuery):
                self.json_format.function_tag_name = query.method_name
            else:
                self.json_format.function_tag_name = None

            if isinstance(return_type, ClientResult):
                if isinstance(return_type.value, ClientValue) or isinstance(return_type.value, ClientObject):
                    return_type = return_type.value
            self.map_json(response.json(), return_type, self.json_format)

    def map_json(self, json, return_type, json_format=None):
        """
        :type json: any
        :type return_type: ClientValue or ClientResult  or ClientObject
        :type json_format: office365.runtime.odata.odata_json_format.ODataJsonFormat
        """
        if json_format is None:
            json_format = self.json_format

        if json and return_type is not None:
            for k, v in self._next_property(json, json_format):
                if isinstance(return_type, ClientResult):
                    return_type.value = v
                elif isinstance(return_type, ClientObjectCollection) and k == json_format.collection_next_tag_name:
                    return_type.next_request_url = v
                else:
                    return_type.set_property(k, v, False)

    def _next_property(self, json, data_format):
        """
        :type json: any
        :type data_format: office365.runtime.odata.odata_json_format.ODataJsonFormat
        """
        if isinstance(data_format, JsonLightFormat):
            json = json.get(data_format.security_tag_name, json)
            json = json.get(data_format.function_tag_name, json)

        if not isinstance(json, dict):
            yield "value", json
        else:
            next_link_url = json.get(self.json_format.collection_next_tag_name, None)
            json = json.get(data_format.collection_tag_name, json)
            if next_link_url:
                yield self.json_format.collection_next_tag_name, next_link_url

            if isinstance(json, list):
                for index, item in enumerate(json):
                    if isinstance(item, dict):
                        item = {k: v for k, v in self._next_property(item, data_format)}
                    yield index, item
            else:
                for name, value in json.items():
                    if isinstance(data_format, JsonLightFormat):
                        is_valid = name != "__metadata" and not (isinstance(value, dict) and "__deferred" in value)
                    else:
                        is_valid = "@odata" not in name

                    if is_valid:
                        if isinstance(value, dict):
                            value = {k: v for k, v in self._next_property(value, data_format)}
                        yield name, value

    def _normalize_payload(self, query):
        """
        Normalizes OData request payload

        :type query: office365.runtime.queries.client_query.ClientQuery
        :rtype: dict
        """
        if isinstance(query.parameter_type, ClientObject) or isinstance(query.parameter_type, ClientValue):
            json = query.parameter_type.to_json(self._json_format)
            if isinstance(query, ServiceOperationQuery) and query.parameter_name is not None:
                return {query.parameter_name: json}
            return json
        elif isinstance(query.parameter_type, dict):
            json = query.parameter_type
            for k, v in json.items():
                if isinstance(v, ClientObject) or isinstance(v, ClientValue):
                    json[k] = v.to_json(self._json_format)
            return json
        else:
            return query.parameter_type



from office365.runtime.client_value import ClientValue


class ContentTypeOrder(ClientValue):
    """The contentTypeOrder resource specifies in which order the Content Type will appear in the selection UI."""
    pass



from office365.entity import Entity


class UserFlowLanguagePage(Entity):
    """
    Determines the user flow language pages that are shown to users during a user flow.
    These language pages include both the default language translations provided by Microsoft,
    or custom pages that can be created to customize the language translations.
    """
    pass



from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.sharing.object_sharing_information import ObjectSharingInformation
from office365.sharepoint.sharing.sharePointSharingSettings import SharePointSharingSettings
from office365.sharepoint.sharing.sharingPermissionInformation import SharingPermissionInformation


class ObjectSharingSettings(BaseEntity):

    @property
    def web_url(self):
        """

        :return: str
        """
        return self.properties.get("WebUrl", None)

    @property
    def access_request_mode(self):
        """

        :return: bool
        """
        return self.properties.get("AccessRequestMode", None)

    @property
    def can_send_email(self):
        """

        :return: bool
        """
        return self.properties.get("CanSendEmail", None)

    @property
    def is_user_site_admin(self):
        """

        :return: bool
        """
        return self.properties.get("IsUserSiteAdmin", None)

    @property
    def object_sharing_information(self):
        return self.properties.get("ObjectSharingInformation",
                                   ObjectSharingInformation(self.context,
                                                            ResourcePath("ObjectSharingInformation",
                                                                         self.resource_path)))

    @property
    def sharepoint_settings(self):
        return self.properties.get("SharePointSettings",
                                   SharePointSharingSettings(self.context,
                                                             ResourcePath("SharePointSettings", self.resource_path)))

    @property
    def sharing_permissions(self):
        return self.properties.get("SharingPermissions",
                                   SharingPermissionInformation(self.context,
                                                                ResourcePath("SharingPermissions", self.resource_path)))



from office365.outlook.mail.message import Message


class EventMessage(Message):
    """A message that represents a meeting request, cancellation, or response (which can be one of the following:
    acceptance, tentative acceptance, or decline)."""
    pass



from office365.base_item import BaseItem
from office365.entity_collection import EntityCollection
from office365.onedrive.columns.column_link import ColumnLink
from office365.onedrive.listitems.item_reference import ItemReference
from office365.runtime.resource_path import ResourcePath


class ContentType(BaseItem):
    """The contentType resource represents a content type in SharePoint. Content types allow you to define a set of
    columns that must be present on every listItem in a list. """

    @property
    def name(self):
        """
        The name of the content type.

        :rtype: str or None
        """
        return self.properties.get('name', None)

    @property
    def description(self):
        """
        The descriptive text for the item.

        :rtype: str or None
        """
        return self.properties.get('description', None)

    @property
    def parent_id(self):
        """The unique identifier of the content type.

        :rtype: str or None
        """
        return self.properties.get('parentId', None)

    @property
    def read_only(self):
        """
        If true, the content type cannot be modified unless this value is first set to false.

        :rtype: bool or None
        """
        return self.properties.get('readOnly', None)

    @property
    def inherited_from(self):
        """
        If this content type is inherited from another scope (like a site),
        provides a reference to the item where the content type is defined.
        """
        return self.properties.get("inheritedFrom", ItemReference())

    @property
    def column_links(self):
        """The collection of columns that are required by this content type"""
        return self.properties.get('columnLinks',
                                   EntityCollection(self.context,
                                                    ColumnLink, ResourcePath("columnLinks", self.resource_path)))



from office365.runtime.queries.create_entity_query import CreateEntityQuery
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.contenttypes.content_type import ContentType


class ContentTypeCollection(BaseEntityCollection):
    """Content Type resource collection"""

    def __init__(self, context, resource_path=None, parent=None):
        super(ContentTypeCollection, self).__init__(context, ContentType, resource_path, parent)

    def get_by_id(self, contentTypeId):
        """
        Returns the content type with the given identifier from the collection.
        If a content type with the given identifier is not found in the collection, the server MUST return null.

        :param str contentTypeId: A hexadecimal value representing the identifier of a content type.
        """
        return ContentType(self.context, ResourcePathServiceOperation("GetById", [contentTypeId], self.resource_path))

    def add(self, content_type_info):
        """Adds a new content type to the collection and returns a reference to the added SP.ContentType.

        :param ContentTypeCreationInformation content_type_info: Specifies properties that is to be used to
            construct the new content type.

        """
        ct = ContentType(self.context)
        self.add_child(ct)
        ct_json = content_type_info.to_json()
        for k, v in ct_json.items():
            if k == "Id":
                ct.set_property(k, {"StringValue": v}, True)
            else:
                ct.set_property(k, v, True)
        qry = CreateEntityQuery(self, ct, ct)
        self.context.add_query(qry)
        return ct

    def add_available_content_type(self, contentTypeId):
        """Adds the specified content type to the content type collection.

        :param str contentTypeId: Specifies the identifier of the content type to be added to the content type
            collection. It MUST exist in the web's available content types.

        """
        ct = ContentType(self.context)
        self.add_child(ct)
        qry = ServiceOperationQuery(self, "AddAvailableContentType", [contentTypeId], None, None, ct)
        self.context.add_query(qry)
        return ct

    def add_existing_content_type(self, content_type):
        """Adds an existing content type to the collection. The name of the given content type MUST NOT be the same
        as any of the content types in the collection. A reference to the SP.ContentType that was added is returned.

        :param ContentType content_type: Specifies the content type to be added to the collection

        """
        pass



from office365.onedrive.internal.queries.upload_session_query import UploadSessionQuery
from office365.outlook.mail.attachment_item import AttachmentItem
from office365.outlook.mail.attachment_type import AttachmentType


class AttachmentUploadQuery(UploadSessionQuery):
    """Create an upload session to allow your app to upload attachments up to the maximum file size. An upload session
    allows your app to upload ranges of the file in sequential API requests, which allows the transfer to be resumed
    if a connection is dropped while the upload is in progress. """

    def create_upload_session(self):
        attachment_item = AttachmentItem(attachment_type=AttachmentType.file, name=self.file_name, size=self.file_size)
        return self.binding_type.create_upload_session(attachment_item)



from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


def create_upload_file_query(file, file_object):
    """
    A Upload file content query

    :type file: office365.sharepoint.files.file.File
    :type file_object: typing.IO
    """
    qry = ServiceOperationQuery(file, "$value")

    def _construct_upload_request(request):
        request.data = file_object.read()
        request.method = HttpMethod.Post
        request.set_header('X-HTTP-Method', 'PUT')
    file.context.before_execute(_construct_upload_request)
    return qry



import copy
from office365.runtime.auth.authentication_context import AuthenticationContext
from office365.runtime.auth.client_credential import ClientCredential
from office365.runtime.auth.providers.saml_token_provider import resolve_base_url
from office365.runtime.auth.user_credential import UserCredential
from office365.runtime.client_runtime_context import ClientRuntimeContext
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.http.request_options import RequestOptions
from office365.runtime.odata.v3.json_light_format import JsonLightFormat
from office365.runtime.odata.v3.batch_request import ODataBatchRequest
from office365.runtime.odata.v3.metadata_level import ODataMetadataLevel
from office365.runtime.odata.odata_request import ODataRequest
from office365.runtime.queries.batch_query import BatchQuery
from office365.runtime.queries.delete_entity_query import DeleteEntityQuery
from office365.runtime.queries.update_entity_query import UpdateEntityQuery
from office365.sharepoint.sites.site import Site
from office365.sharepoint.webs.context_web_information import ContextWebInformation
from office365.sharepoint.webs.web import Web
from office365.runtime.compat import range_or_xrange


class ClientContext(ClientRuntimeContext):
    """SharePoint client context"""

    def __init__(self, base_url, auth_context=None):
        """
        :type base_url: str
        :type auth_context: AuthenticationContext or None
        """
        if base_url.endswith("/"):
            base_url = base_url[:len(base_url) - 1]
        if auth_context is None:
            self._auth_context = AuthenticationContext(url=base_url)
        else:
            self._auth_context = auth_context
        super(ClientContext, self).__init__()
        self.__web = None
        self.__site = None
        self._base_url = base_url
        self._contextWebInformation = None
        self._pendingRequest = ODataRequest(self, JsonLightFormat(ODataMetadataLevel.Verbose))
        self._pendingRequest.beforeExecute += self._build_modification_query

    @staticmethod
    def from_url(abs_url):
        """
        Constructs ClientContext from absolute Url

        :param str abs_url: Absolute Url to resource
        :return: ClientContext
        """
        base_url = resolve_base_url(abs_url)
        ctx = ClientContext(base_url)
        result = Web.get_web_url_from_page_url(ctx, abs_url)

        def _init_context_for_web(resp):
            """
            :type resp: requests.Response
            """
            resp.raise_for_status()
            ctx._base_url = result.value
        ctx.after_execute(_init_context_for_web)
        return ctx

    def with_client_certificate(self, tenant, client_id, thumbprint, cert_path):
        """Creates authenticated SharePoint context via certificate credentials

        :param str tenant: Tenant name
        :param str cert_path: Path to A PEM encoded certificate private key.
        :param str thumbprint: Hex encoded thumbprint of the certificate.
        :param str client_id: The OAuth client id of the calling application.
        """
        self.authentication_context.with_client_certificate(tenant, client_id, thumbprint, cert_path)
        return self

    def with_access_token(self, token_func):
        """
        :type token_func: () -> TokenResponse
        """
        self.authentication_context.register_provider(token_func)
        return self

    def with_user_credentials(self, username, password, allow_ntlm=False, browser_mode=False):
        """
        Assigns credentials

        :type username: str
        :type password: str
        :type allow_ntlm: bool
        :type browser_mode: bool
        """
        self.authentication_context.register_provider(
            UserCredential(username, password),
            allow_ntlm=allow_ntlm,
            browser_mode=browser_mode)
        return self

    def with_credentials(self, credentials):
        """
        Assigns credentials

        :type credentials: UserCredential or ClientCredential
        """
        self.authentication_context.register_provider(credentials)
        return self

    def execute_batch(self, items_per_batch=100):
        """
        Construct and submit a batch request

        :param int items_per_batch: Maximum to be selected for bulk operation
        """
        batch_request = ODataBatchRequest(self)

        def _prepare_batch_request(request):
            self.ensure_form_digest(request)
        batch_request.beforeExecute += _prepare_batch_request

        all_queries = [qry for qry in self.pending_request().next_query()]
        for i in range_or_xrange(0, len(all_queries), items_per_batch):
            queries = all_queries[i:i + items_per_batch]
            batch_request.add_query(BatchQuery(self, queries))
            batch_request.execute_query()

    def build_single_request(self, query):
        """
        :type: office365.runtime.queries.client_query.ClientQuery
        """
        request = super(ClientContext, self).build_single_request(query)
        self._build_modification_query(request)
        return request

    def pending_request(self):
        """
        :return: ODataRequest
        """
        return self._pendingRequest

    def ensure_form_digest(self, request_options):
        """
        :type request_options: RequestOptions
        """
        if not self._contextWebInformation:
            self._contextWebInformation = ContextWebInformation()
            self.request_form_digest()
        request_options.set_header('X-RequestDigest', self._contextWebInformation.FormDigestValue)

    def request_form_digest(self):
        """Request Form Digest"""
        request = RequestOptions("contextInfo")
        request.method = HttpMethod.Post
        response = self.execute_request_direct(request)
        json = response.json()
        json_format = JsonLightFormat()
        json_format.function_tag_name = "GetContextWebInformation"
        self.pending_request().map_json(json, self._contextWebInformation, json_format)

    def clone(self, url, clear_queries=True):
        """
        Creates a clone of ClientContext

        :param bool clear_queries:
        :param str url: Site Url
        :return ClientContext
        """
        ctx = copy.deepcopy(self)
        ctx._base_url = url
        if clear_queries:
            ctx.clear_queries()
        return ctx

    def authenticate_request(self, request):
        self._auth_context.authenticate_request(request)

    def _build_modification_query(self, request):
        """
        Constructs SharePoint specific modification OData request

        :type request: RequestOptions
        """
        query = self.current_query

        if request.method == HttpMethod.Post:
            self.ensure_form_digest(request)
        # set custom SharePoint control headers
        if isinstance(self._pendingRequest.json_format, JsonLightFormat):
            if isinstance(query, DeleteEntityQuery):
                request.ensure_header("X-HTTP-Method", "DELETE")
                request.ensure_header("IF-MATCH", '*')
            elif isinstance(query, UpdateEntityQuery):
                request.ensure_header("X-HTTP-Method", "MERGE")
                request.ensure_header("IF-MATCH", '*')

    @property
    def web(self):
        """Get Web client object"""
        if not self.__web:
            self.__web = Web(self)
        return self.__web

    @property
    def site(self):
        """Get Site client object"""
        if not self.__site:
            self.__site = Site(self)
        return self.__site

    @property
    def base_url(self):
        return self._base_url

    @property
    def authentication_context(self):
        return self._auth_context

    def service_root_url(self):
        return "{0}/_api".format(self.base_url)



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.tenant.administration.hub_site_collection import HubSiteCollection


class SPHubSitesUtility(BaseEntity):

    def __init__(self, context):
        super(SPHubSitesUtility, self).__init__(context, ResourcePath("Microsoft.SharePoint.Portal.SPHubSitesUtility"))

    def get_hub_sites(self):
        hub_sites = HubSiteCollection(self.context)
        qry = ServiceOperationQuery(self, "GetHubSites", None, None, None, hub_sites)
        self.context.add_query(qry)
        return hub_sites



from office365.runtime.client_value import ClientValue


class UserIdInfo(ClientValue):

    def __init__(self):
        """Represents an identity providers unique identifier information"""
        super(UserIdInfo, self).__init__()
        self.NameId = None
        self.NameIdIssuer = None



from office365.directory.identities.userflows.identity_user_flow import IdentityUserFlow


class B2XIdentityUserFlow(IdentityUserFlow):
    """
    Represents a self-service sign up user flow within an Azure Active Directory tenant.

    User flows are used to enable a self-service sign up experience for guest users on an application.
    User flows define the experience the end user sees while signing up, including which identity providers they can
    use to authenticate, along with which attributes are collected as part of the sign up process.
    """

    @property
    def user_flow_type(self):
        """
        The type of user flow. For self-service sign-up user flows,
        the value can only be signUpOrSignIn and cannot be modified after creation.

        :rtype: str or None
        """
        return self.properties.get('userFlowType', None)



class OnlineMeetingProviderType:
    """Specifies the type of a principal."""

    def __init__(self):
        pass

    unknown = 0
    skypeForBusiness = 1
    skypeForConsumer = 2
    teamsForBusiness = 3



from office365.runtime.client_value import ClientValue


class ChangeQuery(ClientValue):
    """Defines a query that is performed against the change log."""

    def __init__(self,
                 alert=False, site=False, web=False, list_=False, item=False, activity=False, file=False, folder=False,
                 user=False, group=False,
                 view=False, content_type=False,
                 add=True, update=True, system_update=True, delete_object=True,
                 role_assignment_add=True, role_assignment_delete=True,
                 change_token_start=None, change_token_end=None, fetch_limit=None):
        """
        :param int fetch_limit:
        :param role_assignment_delete: Specifies whether deleting role assignments is included in the query.
        :param role_assignment_add: Specifies whether adding role assignments is included in the query.
        :param bool item: Gets or sets a value that specifies whether general changes to list items are included
             in the query.
        :param bool delete_object: Gets or sets a value that specifies whether delete changes are included in the query.
        :param bool content_type: Gets or sets a value that specifies whether changes to content types are included
            in the query.
        :param bool alert: Gets or sets a value that specifies whether changes to alerts are included in the query.
        :param bool add: Gets or sets a value that specifies whether add changes are included in the query.
        :param bool view: Gets or sets a value that specifies whether changes to views are included in the query.
        :param bool system_update: Gets or sets a value that specifies whether updates made using the item SystemUpdate
            method are included in the query.
        :param bool update: Gets or sets a value that specifies whether update changes are included in the query.
        :param bool user: Gets or sets a value that specifies whether changes to users are included in the query.
        :param bool folder: Gets or sets value that specifies whether changes to folders are included in the query.
        :param bool file: Gets or sets a value that specifies whether changes to files are included in the query.
        :param change_token_start: office365.sharepoint.changes.changeToken.ChangeToken
        :param change_token_end: office365.sharepoint.changes.changeToken.ChangeToken
        :param bool activity:
        :param bool site: Gets or sets a value that specifies whether changes to site collections
            are included in the query.
        :param bool web: Gets or sets a value that specifies whether changes to Web sites are included in the query.
        :param bool list_: Gets or sets a value that specifies whether changes to lists are included in the query.
        """
        super(ChangeQuery, self).__init__()
        self.Item = item
        self.Alert = alert
        self.ContentType = content_type
        self.Web = web
        self.Site = site
        self.List = list_
        self.Activity = activity
        self.File = file
        self.Folder = folder
        self.User = user
        self.Group = group
        self.View = view
        self.Add = add
        self.Update = update
        self.SystemUpdate = system_update
        self.ChangeTokenStart = change_token_start
        self.ChangeTokenEnd = change_token_end
        self.DeleteObject = delete_object
        self.RoleAssignmentAdd = role_assignment_add
        self.RoleAssignmentDelete = role_assignment_delete
        self.FetchLimit = fetch_limit

    @property
    def entity_type_name(self):
        return 'SP.ChangeQuery'



from office365.entity import Entity


class Presence(Entity):
    """Contains information about a user's presence, including their availability and user activity."""

    @property
    def activity(self):
        """
        :rtype: str or None
        """
        return self.properties.get("activity", None)

    @property
    def availability(self):
        """
        :rtype: str or None
        """
        return self.properties.get("availability", None)



from office365.entity_collection import EntityCollection
from office365.onedrive.listitems.field_value_set import FieldValueSet
from office365.onedrive.versions.base_item_version import BaseItemVersion
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class ListItemVersion(BaseItemVersion):
    """The listItemVersion resource represents a previous version of a ListItem resource."""

    def restore_version(self):
        qry = ServiceOperationQuery(self, "restoreVersion")
        self.context.add_query(qry)
        return self

    @property
    def fields(self):
        """A collection of the fields and values for this version of the list item.

        :rtype: EntityCollection
        """
        return self.get_property('fields',
                                 EntityCollection(self.context, FieldValueSet,
                                                  ResourcePath("fields", self.resource_path)))



from office365.directory.groups.group import Group
from office365.directory.groups.group_profile import GroupProfile
from office365.entity_collection import EntityCollection
from office365.runtime.client_result import ClientResult
from office365.runtime.resource_path import ResourcePath
from office365.teams.team import Team


class TeamCollection(EntityCollection):
    """Team's collection"""

    def __init__(self, context, resource_path=None):
        super(TeamCollection, self).__init__(context, Team, resource_path)

    def __getitem__(self, key):
        """
        :rtype: Team
        """
        return Team(self.context, ResourcePath(key, self.resource_path))

    def get(self):
        """
        :rtype: TeamCollection
        """
        return super(TeamCollection, self).get()

    def get_all(self, include_properties=None):
        """List all teams in Microsoft Teams for an organization"""
        if include_properties is None:
            include_properties = []
        include_properties = include_properties + ["id", "resourceProvisioningOptions"]
        groups = self.context.groups.select(include_properties).get()

        def _process_response(resp):
            for grp in groups:  # type: Group
                if "Team" in grp.properties["resourceProvisioningOptions"]:
                    new_team = Team(self.context, ResourcePath(grp.id, self.resource_path))
                    for k, v in grp.properties.items():
                        new_team.set_property(k, v)
                    self.add_child(new_team)

        self.context.after_execute(_process_response)
        return self

    def create(self, group_name):
        """Provision a new team along with a group.

        :type group_name: str
        :rtype: ClientResult
        """

        grp_properties = GroupProfile(group_name)
        grp_properties.securityEnabled = False
        grp_properties.mailEnabled = True
        grp_properties.groupTypes = ["Unified"]
        target_group = self.context.groups.add(grp_properties)
        result = ClientResult(self.context, Team(self.context))

        def _group_created(resp):
            """
            :type resp: requests.Response
            """
            result.value = target_group.add_team()

        self.context.after_execute(_group_created)
        return result



from office365.reports.report import Report
from office365.runtime.client_result import ClientResult
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


def create_report_query(report_root, report_name, period=None):
    """
    Construct Report query

    :param office365.reports.report_root.ReportRoot report_root: Report container
    :param str report_name: Report name
    :param str period: Specifies the length of time over which the report is aggregated.
        The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
        Dn where n represents the number of days over which the report is aggregated. Required.
    """
    params = {
        "period": period,
    }
    return_type = ClientResult(report_root.context, Report())
    qry = ServiceOperationQuery(report_root, report_name, params, None, None, return_type)

    def _construct_query(request):
        """
        :type request: office365.runtime.http.request_options.RequestOptions
        """
        request.method = HttpMethod.Get

    report_root.context.before_execute(_construct_query)
    return qry



import json

from office365.directory.applications.app_role_assignment import AppRoleAssignmentCollection
from office365.directory.licenses.assigned_license import AssignedLicense
from office365.directory.directory_object import DirectoryObject
from office365.directory.directory_object_collection import DirectoryObjectCollection
from office365.entity_collection import EntityCollection
from office365.onedrive.drives.drive import Drive
from office365.onenote.onenote import Onenote
from office365.outlook.calendar.event import Event
from office365.planner.planner_group import PlannerGroup
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.teams.team import Team


class Group(DirectoryObject):
    """Represents an Azure Active Directory (Azure AD) group, which can be an Office 365 group, or a security group."""

    def renew(self):
        """
        Renews a group's expiration. When a group is renewed, the group expiration is extended by the number
        of days defined in the policy.
        """
        qry = ServiceOperationQuery(self, "renew")
        self.context.add_query(qry)
        return self

    def add_favorite(self):
        """Add the group to the list of the current user's favorite groups. Supported for Microsoft 365 groups only."""
        qry = ServiceOperationQuery(self, "addFavorite")
        self.context.add_query(qry)
        return self

    def remove_favorite(self):
        """
        Remove the group from the list of the current user's favorite groups. Supported for Microsoft 365 groups only.
        """
        qry = ServiceOperationQuery(self, "removeFavorite")
        self.context.add_query(qry)
        return self

    def reset_unseen_count(self):
        """
        Reset the unseenCount of all the posts that the current user has not seen since their last visit.
        Supported for Microsoft 365 groups only.
        """
        qry = ServiceOperationQuery(self, "resetUnseenCount")
        self.context.add_query(qry)
        return self

    def subscribe_by_mail(self):
        """Calling this method will enable the current user to receive email notifications for this group,
        about new posts, events, and files in that group. Supported for Microsoft 365 groups only."""
        qry = ServiceOperationQuery(self, "subscribeByMail")
        self.context.add_query(qry)
        return self

    def unsubscribe_by_mail(self):
        """Calling this method will prevent the current user from receiving email notifications for this group
        about new posts, events, and files in that group. Supported for Microsoft 365 groups only."""
        qry = ServiceOperationQuery(self, "unsubscribeByMail")
        self.context.add_query(qry)
        return self

    def check_member_groups(self, group_ids):
        """Check for membership in the specified list of groups. Returns from the list those groups of which
        the specified group has a direct or transitive membership.

        You can check up to a maximum of 20 groups per request. This function supports Microsoft 365 and other types
        of groups provisioned in Azure AD. Note that Microsoft 365 groups cannot contain groups.
        So membership in a Microsoft 365 group is always direct.

        :type group_ids: list[str]
        """
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "checkMemberGroups", None, group_ids, None, result)
        self.context.add_query(qry)
        return result

    def add_team(self):
        """Create a new team under a group."""
        team = Team(self.context, ResourcePath("team", self.resource_path))
        team._parent_collection = self.parent_collection
        qry = ServiceOperationQuery(self, "team", None, team, None, team)
        self.context.add_query(qry)

        def _construct_create_team_request(request):
            cur_qry = self.context.current_query
            if cur_qry.id == qry.id:
                request.method = HttpMethod.Put
                request.set_header('Content-Type', "application/json")
                request.data = json.dumps(request.data)

        self.context.before_execute(_construct_create_team_request, False)
        return team

    def delete_object(self, permanent_delete=False):
        """
        :param permanent_delete: Permanently deletes the group from directory
        :type permanent_delete: bool

        """
        super(Group, self).delete_object()
        if permanent_delete:
            deleted_item = self.context.directory.deleted_groups[self.id]
            deleted_item.delete_object()
        return self

    @property
    def members(self):
        """Users and groups that are members of this group.

        :rtype: DirectoryObjectCollection
        """
        return self.get_property('members',
                                 DirectoryObjectCollection(self.context, ResourcePath("members", self.resource_path)))

    @property
    def transitive_members(self):
        """
        Get a list of the group's members. A group can have users, devices, organizational contacts,
        and other groups as members. This operation is transitive and returns a flat list of all nested members.

        :rtype: DirectoryObjectCollection
        """
        return self.get_property('transitiveMembers',
                                 DirectoryObjectCollection(self.context,
                                                           ResourcePath("transitiveMembers", self.resource_path)))

    @property
    def transitive_member_of(self):
        """
        Get groups that the group is a member of. This operation is transitive and will also include all groups that
        this groups is a nested member of. Unlike getting a user's Microsoft 365 groups, this returns all
        types of groups, not just Microsoft 365 groups.

        :rtype: DirectoryObjectCollection
        """
        return self.get_property('transitiveMemberOf',
                                 DirectoryObjectCollection(self.context,
                                                           ResourcePath("transitiveMemberOf", self.resource_path)))

    @property
    def owners(self):
        """The owners of the group.

        :rtype: DirectoryObjectCollection
        """
        return self.get_property('owners',
                                 DirectoryObjectCollection(self.context, ResourcePath("owners", self.resource_path)))

    @property
    def drives(self):
        """The group's drives. Read-only.

        :rtype: EntityCollection
        """
        return self.get_property('drives',
                                 EntityCollection(self.context, Drive, ResourcePath("drives", self.resource_path)))

    @property
    def sites(self):
        """The list of SharePoint sites in this group. Access the default site with /sites/root.

        :rtype: SiteCollection
        """
        from office365.onedrive.sites.sites_with_root import SitesWithRoot
        return self.get_property('sites',
                                 SitesWithRoot(self.context, ResourcePath("sites", self.resource_path)))

    @property
    def events(self):
        """Get an event collection or an event."""
        return self.properties.get('events', EntityCollection(self.context, Event,
                                                              ResourcePath("events", self.resource_path)))

    @property
    def app_role_assignments(self):
        """Get an event collection or an appRoleAssignments."""
        return self.properties.get('appRoleAssignments',
                                   AppRoleAssignmentCollection(self.context,
                                                               ResourcePath("appRoleAssignments", self.resource_path)))

    @property
    def onenote(self):
        """Represents the Onenote services available to a group."""
        return self.properties.get('onenote',
                                   Onenote(self.context, ResourcePath("onenote", self.resource_path)))

    @property
    def planner(self):
        """The plannerGroup resource provide access to Planner resources for a group."""
        return self.properties.get('planner',
                                   PlannerGroup(self.context, ResourcePath("planner", self.resource_path)))

    @property
    def assigned_licenses(self):
        return self.properties.get('assignedLicenses', ClientValueCollection(AssignedLicense))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "transitiveMembers": self.transitive_members,
                "transitiveMemberOf": self.transitive_member_of,
                "appRoleAssignments": self.app_role_assignments
            }
            default_value = property_mapping.get(name, None)
        return super(Group, self).get_property(name, default_value)



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.publishing.video_channel_collection import VideoChannelCollection


class VideoServiceManager(BaseEntity):

    def __init__(self, context):
        super(VideoServiceManager, self).__init__(context, ResourcePath("SP.Publishing.VideoServiceManager"))

    def get_channels(self, start_index=0, limit=None):
        return_type = VideoChannelCollection(self.context)
        params = {"startIndex": start_index, "limit": limit}
        qry = ServiceOperationQuery(self, "GetChannels", params, None, None, return_type)
        self.context.add_query(qry)
        return return_type



from office365.runtime.auth.authentication_provider import AuthenticationProvider


class OAuthTokenProvider(AuthenticationProvider):
    """ OAuth token provider for AAD"""

    def __init__(self, token_func=None):
        self._error = None
        self._token_func = token_func
        self._cached_token = None

    def authenticate_request(self, request):
        """
        :type request: office365.runtime.http.request_options.RequestOptions
        """
        if self._cached_token is None:
            self._cached_token = self._token_func()
        request.set_header('Authorization', self._get_authorization_header())

    def _get_authorization_header(self):
        return '{token_type} {access_token}'.format(token_type=self._cached_token.tokenType,
                                                    access_token=self._cached_token.accessToken)

    def get_last_error(self):
        return self._error



from office365.runtime.client_value import ClientValue


class EmailAddress(ClientValue):
    """The name and email address of a contact or message recipient."""

    def __init__(self, address=None, name=None):
        super(EmailAddress, self).__init__()
        self.address = address
        self.name = name



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.search.simpleDataTable import SimpleDataTable


class DocumentCrawlLog(BaseEntity):

    def __init__(self, site):
        super(DocumentCrawlLog, self).__init__(site.context,
                                               ResourcePath(
                                                   "Microsoft.SharePoint.Client.Search.Administration.DocumentCrawlLog"))

    def get_crawled_urls(self, getCountOnly=False):
        """
        Retrieves information about all the contents that were crawled.

        :type getCountOnly: bool"""
        result = SimpleDataTable()
        payload = {
            "getCountOnly": getCountOnly
        }
        qry = ServiceOperationQuery(self, "GetCrawledUrls", None, payload, None, result)
        self.context.add_query(qry)
        return result



from office365.runtime.client_value import ClientValue


class AssignedPlan(ClientValue):
    """
    The assignedPlans property of both the user entity and the organization entity is a collection of assignedPlan.
    """
    pass



from office365.entity import Entity


class IdentityProvider(Entity):
    """
    Represents an Azure Active Directory (Azure AD) identity provider.
    The identity provider can be Microsoft, Google, Facebook, Amazon, LinkedIn, or Twitter.
    The following Identity Providers are in Preview: Weibo, QQ, WeChat, GitHub and any OpenID Connect
    supported providers.
    """
    pass



from office365.runtime.client_value import ClientValue


class ItemReference(ClientValue):
    """The ItemReference resource provides information necessary to address a DriveItem via the API."""

    def __init__(self, name=None, path=None, drive_type=None, site_id=None):
        """
        :param str name: The name of the item being referenced. Read-only.
        :param str site_id: For OneDrive for Business and SharePoint, this property represents the ID of the site
            that contains the parent document library of the driveItem resource. The value is the same as the id
            property of that site resource. It is an opaque string that consists of three identifiers of the site.
            For OneDrive, this property is not populated.
        """
        super(ItemReference, self).__init__()
        self.name = name
        self.path = path
        self.driveType = drive_type



class UserRealmInfo(object):

    def __init__(self, auth_url, federated):
        """

        :type federated: bool
        :type auth_url: str or None
        """
        self.STSAuthUrl = auth_url
        self.IsFederated = federated



class AddFieldOptions:

    def __init__(self):
        pass

    DefaultValue = 0
    AddToDefaultContentType = 1
    AddToNoContentType = 2
    AddToAllContentTypes = 4
    AddFieldInternalNameHint = 8
    AddFieldToDefaultView = 16
    AddFieldCheckDisplayName = 32



import os
import uuid

from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.internal.create_file import create_file_query
from office365.sharepoint.files.file import File
from office365.sharepoint.files.file_creation_information import FileCreationInformation


class UploadSessionQuery(ServiceOperationQuery):
    def __init__(self, files, source_path, chunk_size, chunk_uploaded, chunk_func_args):
        """

        :type files: office365.sharepoint.files.file_collection.FileCollection
        :type source_path: str
        :type chunk_size: int
        :type chunk_uploaded: (int, *)->None
        """

        super(UploadSessionQuery, self).__init__(files)
        self._chunk_size = chunk_size
        self._upload_id = str(uuid.uuid4())
        self._source_path = source_path
        self._chunk_uploaded = chunk_uploaded
        self._chunk_func_args = chunk_func_args
        self._bytes_read = 0
        self._upload_result = None
        self._file_handle = None
        self._file_query = self._build_empty_file_query(files)

    def build_url(self):
        return self._file_query.build_url()

    def _build_empty_file_query(self, files):
        info = FileCreationInformation()
        info.url = self.file_name
        info.overwrite = True

        qry = create_file_query(files, info)
        self.context.after_execute(self._build_upload_session_query)
        return qry

    def _read_next_chunk(self):
        data = self._file_handle.read(self._chunk_size)
        self._bytes_read = self._file_handle.tell()
        return data

    def _has_pending_read(self):
        if self._bytes_read >= self.file_size:
            if self._file_handle is not None:
                self._file_handle.close()
            return False
        if self._file_handle is None:
            self._file_handle = open(self._source_path, 'rb')
        return True

    def _build_upload_session_query(self, response):
        """
        :type response: requests.Response
        """
        response.raise_for_status()

        qry = self.context.current_query
        uploaded_bytes = 0
        if isinstance(qry.return_type, ClientResult):
            uploaded_bytes = int(qry.return_type.value)
        elif isinstance(qry.return_type, File):
            uploaded_bytes = qry.return_type.length

        if callable(self._chunk_uploaded):
            self._chunk_uploaded(uploaded_bytes, *self._chunk_func_args)

        if self._has_pending_read():
            piece = self._read_next_chunk()
            if uploaded_bytes == 0:
                self._upload_result = self.return_type.start_upload(self._upload_id, piece)
            elif uploaded_bytes + len(piece) < self.file_size:
                self._upload_result = self.return_type.continue_upload(self._upload_id, uploaded_bytes, piece)
            else:
                self._return_type = self.return_type.finish_upload(self._upload_id, uploaded_bytes, piece)
            self.context.after_execute(self._build_upload_session_query)

    @property
    def return_type(self):
        """
        :rtype: File
        """
        return self._file_query.return_type

    @property
    def binding_type(self):
        return self._file_query.binding_type

    @property
    def file_name(self):
        return os.path.basename(self._source_path)

    @property
    def file_size(self):
        st = os.stat(self._source_path)
        return st.st_size



from office365.runtime.client_value import ClientValue


class PlannerChecklistItem(ClientValue):
    """
    The plannerChecklistItem resource represents an item in the checklist of a task.
    The checklist on a task is represented by the checklistItems object.
    """
    pass



from office365.runtime.client_value import ClientValue


class SiteCollection(ClientValue):
    """The siteCollection resource provides more information about a site collection."""

    def __init__(self, root=None, hostname=None, data_location_code=None):
        """

        :param office365.onedrive.root.Root root: The hostname for the site collection.
        :param str hostname: The hostname for the site collection.
        :param str data_location_code: The geographic region code for where this site collection resides
        """
        super(SiteCollection, self).__init__()
        self.root = root
        self.hostname = hostname
        self.dataLocationCode = data_location_code







from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.webs.timeZone import TimeZone, TimeZoneCollection


class RegionalSettings(BaseEntity):
    """Represents regional settings that are used on the server that is running SharePoint Server."""

    @property
    def date_format(self):
        """
        Gets the date format that is used on the server.

        :rtype: int or None
        """
        return self.properties.get("DateFormat", None)

    @property
    def locale_id(self):
        """
        Gets the locale identifier in use on the server.

        :rtype: int or None
        """
        return self.properties.get("LocaleId", None)

    @property
    def work_days(self):
        """
        Gets a number that represents the work days of Web site calendars.

        :rtype: int or None
        """
        return self.properties.get("WorkDays", None)

    @property
    def time_zone(self):
        """Gets the time zone that is used on the server."""
        return self.properties.get("TimeZone", TimeZone(self.context, ResourcePath("TimeZone", self.resource_path)))

    @property
    def time_zones(self):
        """Gets the collection of time zones used in a server farm."""
        return self.properties.get("TimeZones",
                                   TimeZoneCollection(self.context, ResourcePath("TimeZones", self.resource_path)))





from office365.runtime.client_value import ClientValue


class FolderView(ClientValue):
    """The FolderView resource provides or sets recommendations on the user-experience of a folder."""

    def __init__(self, sortBy =None, sortOrder=None, viewType=None):
        super(FolderView, self).__init__()
        self.sortBy = sortBy
        self.sortOrder = sortOrder
        self.viewType = viewType



from office365.entity import Entity
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.search.search_query import SearchQuery
from office365.search.search_request import SearchRequest
from office365.search.search_response import SearchResponse


class SearchEntity(Entity):

    def query(self, query_string, entity_types=None):
        """
        Runs the query specified in the request body. Search results are provided in the response.

        :param str query_string: Contains the query terms.
        :param list[str] entity_types: One or more types of resources expected in the response.
            Possible values are: list, site, listItem, message, event, drive, driveItem, externalItem.
        """
        search_request = SearchRequest(query=SearchQuery(query_string), entity_types=entity_types)
        payload = {
            "requests": ClientValueCollection(SearchRequest, [search_request])
        }
        return_type = ClientResult(self.context, ClientValueCollection(SearchResponse))
        qry = ServiceOperationQuery(self, "query", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type




from office365.runtime.client_value import ClientValue


class PickerEntityInformationRequest(ClientValue):

    def __init__(self):
        super(PickerEntityInformationRequest, self).__init__()
        self.Key = None
        self.GroupId = None
        self.PrincipalType = None
        self.EmailAddress = None



from office365.directory.licenses.service_plan_info import ServicePlanInfo
from office365.entity import Entity
from office365.runtime.client_value_collection import ClientValueCollection


class SubscribedSku(Entity):
    """Contains information about a service SKU that a company is subscribed to."""

    @property
    def service_plans(self):
        """Information about the service plans that are available with the SKU. Not nullable"""
        return self.properties.get('servicePlans', ClientValueCollection(ServicePlanInfo))



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.webhooks.subscription import Subscription


class SubscriptionCollection(BaseEntityCollection):
    """Represents a collection of Subscription (WebHook) resources."""

    def __init__(self, context, resource_path=None):
        super(SubscriptionCollection, self).__init__(context, Subscription, resource_path)

    def get_by_id(self, _id):
        """Gets the subscription with the specified ID."""
        return Subscription(self.context, ResourcePathServiceOperation("getById", [_id], self.resource_path))

    def add(self, information):
        """
        :type information: office365.sharepoint.webhooks.subscription_information.SubscriptionInformation
        """
        return_type = Subscription(self.context)
        qry = ServiceOperationQuery(self, "Add", None, information, "parameters", return_type)
        self.context.add_query(qry)
        return return_type

    def remove(self, subscription_id):
        """Removes the subscription with the specified subscriptionId from the collection.

        :param str subscription_id: The ID of the subscription.
        """
        payload = {
            "subscriptionId": subscription_id,
        }
        qry = ServiceOperationQuery(self, "Remove", payload, None, None, None)
        self.context.add_query(qry)
        return self



from office365.runtime.client_value import ClientValue


class SpecialFolder(ClientValue):
    """The SpecialFolder resource groups special folder-related data items into a single structure."""
    pass



from office365.runtime.client_value import ClientValue


class SearchQuery(ClientValue):

    def __init__(self, query_string=None):
        super(SearchQuery, self).__init__()
        self.queryString = query_string



from office365.runtime.client_value import ClientValue


class FieldUrlValue(ClientValue):

    def __init__(self, Url=None, Description=None):
        super(FieldUrlValue, self).__init__()
        self.Url = Url
        self.Description = Description

    @property
    def entity_type_name(self):
        return "SP.FieldUrlValue"



from office365.runtime.client_value import ClientValue


class WikiPageCreationInformation(ClientValue):

    def __init__(self, server_relative_url, content):
        """

        :param str server_relative_url:
        :param str content:
        """
        super(WikiPageCreationInformation, self).__init__()
        self.ServerRelativeUrl = server_relative_url
        self.WikiHtmlContent = content

    @property
    def entity_type_name(self):
        return "SP.Utilities.WikiPageCreationInformation"



from office365.sharepoint.publishing.site_page import SitePage


class TopicSitePage(SitePage):

    @property
    def entity_id(self):
        return self.properties.get("EntityId", None)

    @property
    def entity_type(self):
        return self.properties.get("EntityType", None)



from office365.sharepoint.changes.change import Change


class ChangeFolder(Change):
    """Specifies a change on a folder not contained in a list or document library."""

    @property
    def unique_id(self):
        return self.properties.get("UniqueId", None)

    @property
    def web_id(self):
        return self.properties.get("WebId", None)



from office365.runtime.queries.create_entity_query import CreateEntityQuery
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.lists.list import List


class ListCollection(BaseEntityCollection):
    """Lists collection"""

    def __init__(self, context, resource_path=None):
        super(ListCollection, self).__init__(context, List, resource_path)

    def get_by_title(self, list_title):
        """Retrieve List client object by title

        :type list_title: str
        """
        return List(self.context,
                    ResourcePathServiceOperation("GetByTitle", [list_title], self.resource_path))

    def get_by_id(self, list_id):
        """Retrieve List client object by id

        :type list_id: str
        """
        return List(self.context,
                    ResourcePathServiceOperation("GetById", [list_id], self.resource_path))

    def ensure_events_list(self):
        events_list = List(self.context)
        self.add_child(events_list)
        qry = ServiceOperationQuery(self, "EnsureEventsList", None, None, None, events_list)
        self.context.add_query(qry)
        return events_list

    def ensure_site_assets_library(self):
        """Gets a list that is the default asset location for images or other files, which the users
        upload to their wiki pages."""
        target_list = List(self.context)
        self.add_child(target_list)
        qry = ServiceOperationQuery(self, "ensureSiteAssetsLibrary", None, None, None, target_list)
        self.context.add_query(qry)
        return target_list

    def ensure_site_pages_library(self):
        """Gets a list that is the default location for wiki pages."""
        target_list = List(self.context)
        self.add_child(target_list)
        qry = ServiceOperationQuery(self, "ensureSitePagesLibrary", None, None, None, target_list)
        self.context.add_query(qry)
        return target_list

    def add(self, list_creation_information):
        """Creates a List resource

        :type list_creation_information: office365.sharepoint.lists.list_creation_information.ListCreationInformation
        """
        target_list = List(self.context)
        self.add_child(target_list)
        qry = CreateEntityQuery(self, list_creation_information, target_list)
        self.context.add_query(qry)
        return target_list



from office365.runtime.client_object import ClientObject
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.publishing.portal_health_status import PortalHealthStatus
from office365.sharepoint.tenant.administration.hubsite_properties import HubSiteProperties
from office365.sharepoint.tenant.administration.secondary_administrators_fields_data import \
    SecondaryAdministratorsFieldsData
from office365.sharepoint.tenant.administration.secondary_administrators_info import SecondaryAdministratorsInfo
from office365.sharepoint.tenant.administration.site_properties import SiteProperties
from office365.sharepoint.tenant.administration.site_properties_collection import SitePropertiesCollection
from office365.sharepoint.tenant.administration.sitePropertiesEnumerableFilter import SitePropertiesEnumerableFilter
from office365.sharepoint.tenant.administration.spo_operation import SpoOperation


class Tenant(ClientObject):
    """Represents a SharePoint tenant."""

    def __init__(self, context):
        super(Tenant, self).__init__(
            context,
            ResourcePath("Microsoft.Online.SharePoint.TenantAdministration.Tenant"),
            None,
            "Microsoft.Online.SharePoint.TenantAdministration"
            )

    @staticmethod
    def from_url(admin_site_url):
        """
        :type admin_site_url: str
        """
        from office365.sharepoint.client_context import ClientContext
        admin_client = ClientContext(admin_site_url)
        return Tenant(admin_client)

    def get_lock_state_by_id(self, site_id):
        return self._sites.get_lock_state_by_id(site_id)

    def hub_sites(self, siteUrl):
        pass

    def check_tenant_licenses(self, licenses):
        """
        Checks whether a tenant has the specified licenses.

        :param list[str] licenses: The list of licenses to check for.
        :return:
        """
        result = ClientResult(self.context)
        params = ClientValueCollection(str, licenses)
        qry = ServiceOperationQuery(self, "CheckTenantLicenses", None, params, "licenses", result)
        self.context.add_query(qry)
        return result

    def get_site_health_status(self, source_url):
        """
        :type source_url: str
        """
        result = ClientResult(self.context, PortalHealthStatus())
        params = {"sourceUrl": source_url}
        qry = ServiceOperationQuery(self, "GetSiteHealthStatus", None, params, None, result)
        self.context.add_query(qry)
        return result

    def get_site_secondary_administrators(self, site_id):
        """
        Gets site collection administrators

        :type site_id: str
        """
        result = ClientResult(self.context, ClientValueCollection(SecondaryAdministratorsInfo))
        payload = SecondaryAdministratorsFieldsData(site_id)
        qry = ServiceOperationQuery(self, "GetSiteSecondaryAdministrators", None, payload,
                                    "secondaryAdministratorsFieldsData", result)
        self.context.add_query(qry)
        return result

    def set_site_secondary_administrators(self, site_id, emails=None, names=None):
        """
        Sets site collection administrators

        :type names: list[str] or None
        :type emails: list[str]
        :type site_id: str
        """
        payload = SecondaryAdministratorsFieldsData(site_id, emails, names)
        qry = ServiceOperationQuery(self, "SetSiteSecondaryAdministrators", None, payload,
                                    "secondaryAdministratorsFieldsData", None)
        self.context.add_query(qry)
        return self

    def register_hub_site(self, site_url):
        """
        Registers an existing site as a hub site.

        :param str site_url:
        :return:
        """
        return_type = HubSiteProperties(self.context)
        params = {"siteUrl": site_url}
        qry = ServiceOperationQuery(self, "RegisterHubSite", None, params, None, return_type)
        self.context.add_query(qry)
        return return_type

    def unregister_hub_site(self, site_url):
        """
        Unregisters a hub site so that it is no longer a hub site.

        :param str site_url:
        :return:
        """
        params = {"siteUrl": site_url}
        qry = ServiceOperationQuery(self, "UnregisterHubSite", None, params, None, None)
        self.context.add_query(qry)
        return self

    def create_site(self, site_create_props):
        """Queues a site collection for creation with the specified properties.

        :param SiteCreationProperties site_create_props:
        A SiteCreationProperties object that contains the initial properties
        of the new site collection.
        """
        result = SpoOperation(self.context)
        qry = ServiceOperationQuery(self, "CreateSite", None, site_create_props, "siteCreationProperties", result)
        self.context.add_query(qry)
        return result

    def remove_site(self, site_url):
        """Deletes the site with the specified URL

        :param str site_url: A string representing the URL of the site.
        """
        result = SpoOperation(self.context)
        qry = ServiceOperationQuery(self, "removeSite", [site_url], None, None, result)
        self.context.add_query(qry)
        return result

    def remove_deleted_site(self, site_url):
        """Permanently removes the specified deleted site from the recycle bin.

        :param str site_url: A string representing the URL of the site.
        """
        result = SpoOperation(self.context)
        qry = ServiceOperationQuery(self, "RemoveDeletedSite", [site_url], None, None, result)
        self.context.add_query(qry)
        return result

    def restore_deleted_site(self, site_url):
        """Restores deleted site with the specified URL

        :param str site_url: A string representing the URL of the site.
        """
        result = SpoOperation(self.context)
        qry = ServiceOperationQuery(self, "RestoreDeletedSite", [site_url], None, None, result)
        self.context.add_query(qry)
        return result

    def get_site_properties_by_url(self, url, include_detail):
        """

        :param str url: A string that represents the site URL.
        :param bool include_detail: A Boolean value that indicates whether to include all of the SPSite properties.
        """
        site_props = SiteProperties(self.context)
        self._sites.add_child(site_props)
        payload = {
            'url': url,
            'includeDetail': include_detail
        }
        qry = ServiceOperationQuery(self, "getSitePropertiesByUrl", None, payload, None, site_props)
        self.context.add_query(qry)
        return site_props

    def get_site_properties_from_sharepoint_by_filters(self, _filter, start_index=0, include_detail=False):
        """

        :param bool include_detail:
        :param int start_index:
        :param str _filter:
        """
        site_props_col = SitePropertiesCollection(self.context)
        qry = ServiceOperationQuery(self, "getSitePropertiesFromSharePointByFilters",
                                    None,
                                    SitePropertiesEnumerableFilter(_filter, start_index, include_detail),
                                    "speFilter",
                                    site_props_col)
        self.context.add_query(qry)
        return site_props_col

    def connect_site_to_hub_site_by_id(self, site_url, hub_site_id):
        """

        :param str site_url:
        :param str hub_site_id:
        :return:
        """
        params = {
            "siteUrl": site_url,
            "hubSiteId": hub_site_id
        }
        qry = ServiceOperationQuery(self, "ConnectSiteToHubSiteById", None, params, None, None)
        self.context.add_query(qry)
        return self

    @property
    def root_site_url(self):
        """

        :rtype: str or None
        """
        return self.properties.get('RootSiteUrl', None)

    @property
    def _sites(self):
        """Gets a collection of sites."""
        return self.properties.get('sites',
                                   SitePropertiesCollection(self.context, ResourcePath("sites", self.resource_path)))



import base64

from office365.entity_collection import EntityCollection
from office365.onedrive.shares.shared_drive_item import SharedDriveItem
from office365.runtime.resource_path import ResourcePath


def _url_to_shared_token(url):
    value = base64.b64encode(url.encode("ascii")).decode("ascii")
    if value.endswith("="):
        value = value[:-1]
    return "u!" + value.replace('/', '_').replace('+', '-')


class SharesCollection(EntityCollection):

    def __init__(self, context, resource_path=None):
        super(SharesCollection, self).__init__(context, SharedDriveItem, resource_path)

    def by_url(self, url):
        """
        Address shared item by absolute url

        :type url: str
        """
        return SharedDriveItem(self.context, ResourcePath(_url_to_shared_token(url), self.resource_path))



from office365.directory.directory_object import DirectoryObject
from office365.runtime.client_value_collection import ClientValueCollection


class ExtensionProperty(DirectoryObject):
    """
    Represents a directory extension that can be used to add a custom property to directory objects without
    requiring an external data store. For example, if an organization has a line of business (LOB) application
    that requires a Skype ID for each user in the directory, Microsoft Graph can be used to register a new property
    named skypeId on the directory’s User object, and then write a value to the new property for a specific user.
    """

    @property
    def name(self):
        """
        Name of the extension property.

        :rtype: str
        """
        return self.properties.get("name", None)

    @property
    def app_display_name(self):
        """
        Display name of the application object on which this extension property is defined. Read-only.

        :rtype: str
        """
        return self.properties.get("appDisplayName", None)

    @property
    def data_type(self):
        """
        Specifies the data type of the value the extension property can hold. Following values are supported.
            Binary - 256 bytes maximum
            Boolean
            DateTime - Must be specified in ISO 8601 format. Will be stored in UTC.
            Integer - 32-bit value.
            LargeInteger - 64-bit value.
            String - 256 characters maximum

        :rtype: str
        """
        return self.properties.get("dataType", None)

    @property
    def target_objects(self):
        """
        Following values are supported. Not nullable.
        User
        Group
        Organization
        Device
        Application
        """
        return self.properties.get("targetObjects", ClientValueCollection(str))



from office365.entity import Entity


class GroupLifecyclePolicy(Entity):
    """
    Represents a lifecycle policy for a Microsoft 365 group. A group lifecycle policy allows administrators
    to set an expiration period for groups. For example, after 180 days, a group expires.
    When a group reaches its expiration, owners of the group are required to renew their group within a time interval
    defined by the administrator. Once renewed, the group expiration is extended by the number of days defined
    in the policy. For example, the group's new expiration is 180 days after renewal.
    If the group is not renewed, it expires and is deleted.
    The group can be restored within a period of 30 days from deletion.
    """
    pass



from abc import ABCMeta, abstractmethod


class AuthenticationProvider(object):
    """ Base token provider"""
    __metaclass__ = ABCMeta

    @abstractmethod
    def authenticate_request(self, request):
        """

        :type request: office365.runtime.http.request_options.RequestOptions
        """
        pass



from office365.base_item import BaseItem
from office365.onedrive.columns.calculated_column import CalculatedColumn
from office365.onedrive.columns.choice_column import ChoiceColumn
from office365.onedrive.columns.default_column_value import DefaultColumnValue
from office365.onedrive.columns.geolocation_column import GeolocationColumn
from office365.onedrive.columns.lookup_column import LookupColumn
from office365.onedrive.columns.number_column import NumberColumn
from office365.onedrive.columns.person_or_group_column import PersonOrGroupColumn
from office365.onedrive.columns.text_column import TextColumn


class ColumnDefinition(BaseItem):

    @property
    def enforce_unique_values(self):
        """
        If true, no two list items may have the same value for this column.

        :rtype: bool or None
        """
        return self.properties.get('enforceUniqueValues', None)

    @property
    def indexed(self):
        """Specifies whether the column values can used for sorting and searching."""
        return self.properties.get('indexed', None)

    @property
    def column_group(self):
        """For site columns, the name of the group this column belongs to. Helps organize related columns."""
        return self.properties.get('columnGroup', None)

    @property
    def geolocation(self):
        """This column stores a geolocation."""
        return self.get_property('geolocation', GeolocationColumn())

    @property
    def calculated(self):
        """This column's data is calculated based on other columns."""
        return self.get_property('calculated', CalculatedColumn())

    @property
    def choice(self):
        """This column stores data from a list of choices."""
        return self.get_property('choice', ChoiceColumn())

    @property
    def person_or_group(self):
        """This column stores Person or Group values."""
        return self.get_property('personOrGroup', PersonOrGroupColumn())

    @property
    def text(self):
        """This column stores text values."""
        return self.get_property('text', TextColumn())

    @property
    def number(self):
        """This column stores number values."""
        return self.get_property('number', NumberColumn())

    @property
    def lookup(self):
        """This column's data is looked up from another source in the site."""
        return self.get_property('lookup', LookupColumn())

    @property
    def default_value(self):
        """The default value for this column."""
        return self.get_property('defaultValue', DefaultColumnValue())



from office365.entity_collection import EntityCollection
from office365.onenote.notebooks.copy_notebook_model import CopyNotebookModel
from office365.onenote.notebooks.notebook import Notebook
from office365.onenote.notebooks.recent_notebook import RecentNotebook
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


class NotebookCollection(EntityCollection):

    def __init__(self, context, resource_path=None):
        super(NotebookCollection, self).__init__(context, Notebook, resource_path)

    def get_notebook_from_web_url(self, web_url):
        """
        Retrieve the properties and relationships of a notebook object by using its URL path.
        The location can be user notebooks on Microsoft 365, group notebooks,
        or SharePoint site-hosted team notebooks on Microsoft 365.

        :param str web_url: The URL path of the notebook to retrieve. It can also contain a "onenote:" prefix.
        """
        result = ClientResult(self.context, CopyNotebookModel())
        params = {"webUrl": web_url}
        qry = ServiceOperationQuery(self, "getNotebookFromWebUrl", params, None, None, result)
        self.context.add_query(qry)
        return result

    def get_recent_notebooks(self, include_personal_notebooks=True):
        """Get a list of recentNotebook instances that have been accessed by the signed-in user.

        :param bool include_personal_notebooks: Include notebooks owned by the user. Set to true to include notebooks
            owned by the user; otherwise, set to false. If you don't include the includePersonalNotebooks parameter,
            your request will return a 400 error response.
        """

        result = ClientResult(self.context, ClientValueCollection(RecentNotebook))
        params = {"includePersonalNotebooks": include_personal_notebooks}
        qry = ServiceOperationQuery(self, "getRecentNotebooks", params, None, None, result)
        self.context.add_query(qry)

        def _construct_request(request):
            request.method = HttpMethod.Get

        self.context.before_execute(_construct_request)
        return result



from office365.runtime.client_value import ClientValue


class CalculatedColumn(ClientValue):
    """The calculatedColumn on a columnDefinition resource indicates that the column's
    data is calculated based on other columns in the site."""
    pass



from office365.onenote.entity_base_model import OnenoteEntityBaseModel


class OnenoteEntitySchemaObjectModel(OnenoteEntityBaseModel):
    """This is a base type for OneNote entities."""

    @property
    def created_datetime(self):
        """
        The date and time when the page was created. The timestamp represents date and time information using
        ISO 8601 format and is always in UTC time. For example, midnight UTC on Jan 1, 2014 is 2014-01-01T00:00:00Z.

        :rtype: str or None
        """
        return self.properties("createdDateTime", None)



from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity


class PushNotificationSubscriber(BaseEntity):
    """Represents a push notification subscriber over a site."""

    @property
    def custom_args(self):
        """Gets the custom arguments specified by the app.

        :rtype: str
        """
        return self.properties.get("CustomArgs", None)

    @property
    def user(self):
        """Gets the SharePoint user who created this subscriber."""
        from office365.sharepoint.principal.user import User
        return self.properties.get("User", User(self.context, ResourcePath("user", self.resource_path)))



from office365.directory.directory_object import DirectoryObject
from office365.entity_collection import EntityCollection
from office365.runtime.client_result import ClientResult
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


class DirectoryObjectCollection(EntityCollection):
    """DirectoryObject's collection"""

    def __init__(self, context, resource_path=None):
        super(DirectoryObjectCollection, self).__init__(context, DirectoryObject, resource_path)

    def get(self):
        """
        :rtype: DirectoryObjectCollection
        """
        return super(DirectoryObjectCollection, self).get()

    def __getitem__(self, key):
        """
        :type key: int or str
        :rtype: DirectoryObject
        """
        return super(DirectoryObjectCollection, self).__getitem__(key)

    def get_by_ids(self, ids):
        """
        Returns the directory objects specified in a list of IDs.

        :type ids: list[str]
        """
        result = ClientResult(self.context)
        params = {
            "ids": ids
        }
        qry = ServiceOperationQuery(self, "getByIds", params, None, None, result)
        self.context.add_query(qry)
        return result

    def add(self, user_id):
        """
        Add a user to the group.

        :type user_id: str
        """
        payload = {
            "@odata.id": "https://graph.microsoft.com/v1.0/users/{0}".format(user_id)
        }
        qry = ServiceOperationQuery(self, "$ref", None, payload)
        self.context.add_query(qry)
        return self

    def get_available_extension_properties(self, is_synced_from_on_premises=None):
        """
        Return all or a filtered list of the directory extension properties that have been registered in a directory.
        The following entities support extension properties: user, group, organization, device, application,
        and servicePrincipal.
        """
        from office365.directory.extensions.extension_property import ExtensionProperty
        return_type = EntityCollection(self.context, ExtensionProperty, self.context.directory_objects.resource_path)
        payload = {
            "isSyncedFromOnPremises": is_synced_from_on_premises
        }
        qry = ServiceOperationQuery(self, "getAvailableExtensionProperties", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def remove(self, user_id):
        """Remove a user from the group."""

        qry = ServiceOperationQuery(self, "{0}/$ref".format(user_id))
        self.context.add_query(qry)

        def _construct_remove_user_request(request):
            """
            :type request: RequestOptions
            """
            request.method = HttpMethod.Delete
        self.context.before_execute(_construct_remove_user_request)
        return self

    def validate_properties(self, entity_type=None, display_name=None, mail_nickname=None, on_behalf_of_userid=None):
        """
        Validate that a Microsoft 365 group's display name or mail nickname complies with naming policies.
        Clients can use this API to determine whether a display name or mail nickname is valid before trying to
        create a Microsoft 365 group. To validate the properties of an existing group, use the group:
        validateProperties function.

        :param str entity_type: Group is the only supported entity type.
        :param str display_name: The display name of the group to validate. The property is not individually required.
             However, at least one property (displayName or mailNickname) is required.
        :param str mail_nickname: The mail nickname of the group to validate.
             The property is not individually required. However, at least one property (displayName or mailNickname)
             is required.
        :param str on_behalf_of_userid: The ID of the user to impersonate when calling the API. The validation results are for the
            onBehalfOfUserId's attributes and roles.
        """
        payload = {
            "entityType": entity_type,
            "displayName": display_name,
            "mailNickname": mail_nickname,
            "onBehalfOfUserId": on_behalf_of_userid
        }
        qry = ServiceOperationQuery(self, "validateProperties", None, payload)
        self.context.add_query(qry)
        return self



from office365.runtime.client_value import ClientValue
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.base_entity_collection import BaseEntityCollection


class EventReceiverDefinitionCreationInformation(ClientValue):
    """Represents the properties that can be set when creating a client-side event receiver definition."""
    pass


class EventReceiverDefinition(BaseEntity):
    """Abstract base class that defines general properties of an event receiver for list items, lists,
    websites, and workflows."""

    @property
    def receiver_url(self):
        """Gets the URL of the receiver for the event.

        :rtype: str or None
        """
        return self.properties.get('ReceiverUrl', None)


class EventReceiverDefinitionCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None, parent=None):
        super(EventReceiverDefinitionCollection, self).__init__(context, EventReceiverDefinition, resource_path, parent)



from office365.sharepoint.fields.field_lookup import FieldLookup


class FieldUser(FieldLookup):
    """Specifies a field that contains a user."""

    @property
    def presence(self):
        """Gets a value that specifies whether presence is enabled on the field.

        :rtype: bool or None
        """
        return self.properties.get('Presence', None)

    @presence.setter
    def presence(self, val):
        """Sets a value that specifies whether presence is enabled on the field.
        """
        self.properties.get('Presence', val)

    @property
    def allow_display(self):
        """Gets a value that specifies whether to display the name of the user in a survey list.

        :rtype: bool or None
        """
        return self.properties.get('AllowDisplay', None)

    @allow_display.setter
    def allow_display(self, val):
        """Sets a value that specifies whether to display the name of the user in a survey list.
        """
        self.properties.get('AllowDisplay', val)

    @property
    def selection_group(self):
        """Gets a value that specifies the identifier of the SharePoint group whose members can be selected
            as values of the field.

        :rtype: int or None
        """
        return self.properties.get('SelectionGroup', None)

    @selection_group.setter
    def selection_group(self, val):
        """Sets a value that specifies the identifier of the SharePoint group whose members can be selected as
            values of the field.
        """
        self.properties.get('SelectionGroup', val)



from office365.runtime.queries.client_query import ClientQuery


class DeleteEntityQuery(ClientQuery):
    def __init__(self, entity_to_delete):
        """
        Delete entity query

        :type entity_to_delete: office365.runtime.client_object.ClientObject
        """
        super(DeleteEntityQuery, self).__init__(entity_to_delete.context, entity_to_delete)



from office365.runtime.resource_path import ResourcePath


class RootResourcePath(ResourcePath):
    """Resource path for OneDrive path-based addressing"""

    def __init__(self, parent):
        super(RootResourcePath, self).__init__("root", parent)





from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity import BaseEntity


class ListTemplate(BaseEntity):

    def __init__(self, context, resource_path=None):
        """
        Represents a list definition or a list template, which defines the fields and views for a list.
            List definitions are contained in files within
            \\Program Files\\Common Files\\Microsoft Shared\\Web Server Extensions\\12\\TEMPLATE\\FEATURES,
            but list templates are created through the user interface or through the object model when a list is
            saved as a template.
            Use the Web.ListTemplates property (section 3.2.5.143.1.2.13) to return a ListTemplateCollection
            (section 3.2.5.92) for a site collection. Use an indexer to return a single list definition or
            list template from the collection.

        """
        super(ListTemplate, self).__init__(context, resource_path)

    def get_global_schema_xml(self):
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "GetGlobalSchemaXml", None, None, None, result)
        self.context.add_query(qry)
        return result

    @property
    def internal_name(self):
        """Gets a value that specifies the identifier for the list template.
        :rtype: str or None
        """
        return self.properties.get('InternalName', None)

    def set_property(self, name, value, persist_changes=True):
        super(ListTemplate, self).set_property(name, value, persist_changes)
        if self._resource_path is None:
            if name == "Name":
                self._resource_path = ResourcePathServiceOperation(
                    "GetByName", [value], self._parent_collection.resource_path)
        return self



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.attachments.attachmentfile import AttachmentFile
from office365.sharepoint.attachments.attachmentfile_creation_information import AttachmentfileCreationInformation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.files.file import File


class AttachmentFileCollection(BaseEntityCollection):
    """Represents a collection of AttachmentFile resources."""

    def __init__(self, context, resource_path=None):
        super(AttachmentFileCollection, self).__init__(context, AttachmentFile, resource_path)

    def get(self):
        """
        :rtype: AttachmentFileCollection
        """
        return super(AttachmentFileCollection, self).get()

    def add(self, attachment_file_information):
        """
        Creates an attachment
        :type attachment_file_information: AttachmentfileCreationInformation
        """
        if isinstance(attachment_file_information, dict):
            attachment_file_information = AttachmentfileCreationInformation(
                attachment_file_information.get('filename'),
                attachment_file_information.get('content')
            )

        target_file = File(self.context)
        self.add_child(target_file)
        qry = ServiceOperationQuery(self,
                                    "add",
                                    {
                                        "filename": attachment_file_information.filename,
                                    },
                                    attachment_file_information.content,
                                    None,
                                    target_file)
        self.context.add_query(qry)
        return target_file

    def add_using_path(self):
        pass

    def get_by_filename(self, filename):
        """Retrieve Attachment file object by filename

        :type filename: str
        """
        return AttachmentFile(context=self.context,
                              resource_path=ResourcePathServiceOperation("GetByFileName", [filename],
                                                                         self.resource_path),
                              parent_collection=self)



from office365.runtime.odata.odata_json_format import ODataJsonFormat


class V4JsonFormat(ODataJsonFormat):
    """V4 JSON format"""

    def __init__(self, metadata):
        super(V4JsonFormat, self).__init__(metadata)
        """The IEEE754Compatible format parameter indicates that the service MUST serialize Edm.Int64 and
        Edm.Decimal numbers as strings."""
        self.IEEE754Compatible = False
        """"""
        self.streaming = False
        self.collection_tag_name = "value"
        self.metadata_type_tag_name = "@odata.type"

    def get_media_type(self):
        return "application/json;odata.metadata={0};odata.streaming={1};IEEE754Compatible={2}" \
            .format(self.metadata, self.streaming, self.IEEE754Compatible)



from office365.runtime.client_value import ClientValue


class RecentNotebook(ClientValue):
    """A recently accessed OneNote notebook. A recentNotebook is similar to a notebook but has fewer properties."""
    pass



from office365.sharepoint.base_entity import BaseEntity


class FieldLink(BaseEntity):
    """Specifies a reference to a field or field definition for a content type."""

    @property
    def id(self):
        """Gets a value that specifies the GUID of the FieldLink.

        :rtype: str or None
        """
        return self.properties.get('Id', None)



from office365.runtime.client_value import ClientValue


class UsageInfo(ClientValue):
    """
    Provides fields used to access information regarding site collection usage.
    """
    def __init__(self, bandwidth=None, discussion_storage=None, visits=None):
        """
        :param long bandwidth: Contains the cumulative bandwidth used by the site collection on the previous day or
            on the last day that log files were processed, which is tracked by usage analysis code.
        :param long discussion_storage: Contains the amount of storage, identified in bytes,
            used by Web discussion data in the site collection.
        :param long visits: Contains the cumulative number of visits to the site collection,
            which is tracked by the usage analysis code.
        """
        super(UsageInfo, self).__init__()
        self.Bandwidth = bandwidth
        self.DiscussionStorage = discussion_storage
        self.Visits = visits



from office365.runtime.queries.create_entity_query import CreateEntityQuery
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.principal.group import Group


class GroupCollection(BaseEntityCollection):
    """Represents a collection of Group resources."""
    def __init__(self, context, resource_path=None):
        super(GroupCollection, self).__init__(context, Group, resource_path)

    def add(self, group_creation_information):
        """Creates a Group resource

        :type group_creation_information: any
        """
        group = Group(self.context)
        self.add_child(group)
        qry = CreateEntityQuery(self, group_creation_information, group)
        self.context.add_query(qry)
        return group

    def get_by_id(self, group_id):
        """Returns the list item with the specified list item identifier.

        :type group_id: str
        """
        group = Group(self.context,
                      ResourcePathServiceOperation("getbyid", [group_id], self.resource_path))
        return group

    def get_by_name(self, group_name):
        """Returns a cross-site group from the collection based on the name of the group.

        :type group_name: str
        """
        return Group(self.context,
                     ResourcePathServiceOperation("getByName", [group_name], self.resource_path))

    def remove_by_id(self, group_id):
        """Removes the group with the specified member ID from the collection.

        :type group_id: str
        """
        qry = ServiceOperationQuery(self, "removeById", [group_id])
        self.context.add_query(qry)

    def remove_by_login_name(self, group_name):
        """Removes the cross-site group with the specified name from the collection.

        :type group_name: str
        """
        qry = ServiceOperationQuery(self, "removeByLoginName", [group_name])
        self.context.add_query(qry)



from office365.sharepoint.base_entity import BaseEntity


class Form(BaseEntity):
    """A form provides a display and editing interface for a single list item."""

    @property
    def form_type(self):
        """
        Gets the type of the form.

        :rtype: str or None
        """
        return self.properties.get("FormType", None)

    @property
    def server_relative_url(self):
        """
        Gets the server-relative URL of the form.

        :rtype: str or None
        """
        return self.properties.get("ServerRelativeUrl", None)



from office365.runtime.client_value import ClientValue


class Thumbnail(ClientValue):
    """
    The thumbnail resource type represents a thumbnail for an image, video, document,
        or any item that has a bitmap representation.
    """
    pass



from office365.outlook.calendar.meeting_time_suggestion import MeetingTimeSuggestion
from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection


class MeetingTimeSuggestionsResult(ClientValue):
    """
    A collection of meeting suggestions if there is any, or the reason if there isn't.
    """
    def __init__(self, meetingTimeSuggestions=None,
                 emptySuggestionsReason=None):
        """

        :param ClientValueCollection(MeetingTimeSuggestion) meetingTimeSuggestions: An array of meeting suggestions.
        :param str emptySuggestionsReason: A reason for not returning any meeting suggestions.
            The possible values are: attendeesUnavailable, attendeesUnavailableOrUnknown, locationsUnavailable,
            organizerUnavailable, or unknown. This property is an empty string if the meetingTimeSuggestions property
            does include any meeting suggestions.
        """
        super(MeetingTimeSuggestionsResult, self).__init__()
        self.meetingTimeSuggestions = ClientValueCollection(MeetingTimeSuggestion) if meetingTimeSuggestions is None \
            else meetingTimeSuggestions
        self.emptySuggestionsReason = emptySuggestionsReason



from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.onenote.notebooks.notebook_collection import NotebookCollection
from office365.onenote.operations.onenote_operation import OnenoteOperation
from office365.onenote.pages.page import OnenotePage
from office365.onenote.resources.resource import OnenoteResource
from office365.onenote.sectiongroups.section_group import SectionGroup
from office365.onenote.sections.section import OnenoteSection
from office365.runtime.resource_path import ResourcePath


class Onenote(Entity):

    @property
    def notebooks(self):
        """Retrieve a list of notebook objects.

        :rtype: NotebookCollection
        """
        return self.get_property('notebooks',
                                 NotebookCollection(self.context, ResourcePath("notebooks", self.resource_path)))

    @property
    def operations(self):
        """Retrieve a list of OneNote operations.

        :rtype: EntityCollection
        """
        return self.get_property('operations',
                                 EntityCollection(self.context, OnenoteOperation,
                                                  ResourcePath("operations", self.resource_path)))

    @property
    def pages(self):
        """Retrieve a list of page objects.

        :rtype: EntityCollection
        """
        return self.get_property('pages',
                                 EntityCollection(self.context, OnenotePage,
                                                  ResourcePath("pages", self.resource_path)))

    @property
    def resources(self):
        """Retrieve a list of Resources objects from the specified notebook.

        :rtype: EntityCollection
        """
        return self.get_property('resources',
                                 EntityCollection(self.context, OnenoteResource,
                                                  ResourcePath("resources", self.resource_path)))

    @property
    def sections(self):
        """Retrieve a list of onenoteSection objects from the specified notebook.

        :rtype: EntityCollection
        """
        return self.get_property('sections',
                                 EntityCollection(self.context, OnenoteSection,
                                                  ResourcePath("sections", self.resource_path)))

    @property
    def section_groups(self):
        """Retrieve a list of onenoteSection objects from the specified notebook.

        :rtype: EntityCollection
        """
        return self.get_property('sectionGroups',
                                 EntityCollection(self.context, SectionGroup,
                                                  ResourcePath("sectionGroups", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "sectionGroups": self.section_groups
            }
            default_value = property_mapping.get(name, None)
        return super(Onenote, self).get_property(name, default_value)



from office365.outlook.calendar.schedule_item import ScheduleItem
from office365.outlook.calendar.working_hours import WorkingHours
from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection


class ScheduleInformation(ClientValue):
    """Represents the availability of a user, distribution list, or resource (room or equipment)
    for a specified time period."""

    def __init__(self, scheduleId=None, scheduleItems=None, availabilityView=None, error=None, workingHours=None):
        """

        :param WorkingHours workingHours: The days of the week and hours in a specific time zone that the user works.
             These are set as part of the user's mailboxSettings.
        :param str error: Error information from attempting to get the availability of the user, distribution list,
             or resource.
        :param str availabilityView: Represents a merged view of availability of all the items in scheduleItems.
             The view consists of time slots. Availability during each time slot is indicated with:
             0= free, 1= tentative, 2= busy, 3= out of office, 4= working elsewhere.
        :param ClientValueCollection(ScheduleItem) scheduleItems: Contains the items that describe the availability
            of the user or resource.
        :param str scheduleId: An SMTP address of the user, distribution list, or resource, identifying an instance
            of scheduleInformation.
        """
        super(ScheduleInformation, self).__init__()
        self.scheduleItems = ClientValueCollection(ScheduleItem) if scheduleItems is None else scheduleItems
        self.scheduleId = scheduleId
        self.availabilityView = availabilityView
        self.error = error
        self.workingHours = WorkingHours() if workingHours is None else workingHours



from office365.runtime.client_value import ClientValue


class FileSystemInfo(ClientValue):
    """The FileSystemInfo resource contains properties that are reported by the device's local file system for the
    local version of an item. """

    def __init__(self):
        super(FileSystemInfo, self).__init__()
        self.createdDateTime = None
        self.lastAccessedDateTime = None
        self.lastModifiedDateTime = None



from office365.runtime.client_value import ClientValue


class FieldLookupValue(ClientValue):

    def __init__(self, lookup_id, lookup_value=None):
        """Specifies the value of a lookup for a fields within a list item.

        :param int lookup_id: Gets or sets the identifier (ID) of the list item that this instance of the lookup
        fields is referring to.
        :param str or None lookup_value: Gets a summary of the list item that this instance
        of the lookup fields is referring to.

        """
        super(FieldLookupValue, self).__init__()
        self.LookupId = lookup_id
        self.LookupValue = lookup_value

    @property
    def entity_type_name(self):
        return "SP.FieldLookupValue"



from office365.runtime.client_value import ClientValue


class KeyCredential(ClientValue):
    """
    Contains a key credential associated with an application .
    The keyCredentials property of the application entity is a collection of keyCredential.
    """




from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.views.view import View


class ViewCollection(BaseEntityCollection):
    """Represents a collection of View resources."""

    def __init__(self, context, resource_path=None, parent_list=None):
        """

        :type parent_list: office365.sharepoint.list.List or None
        """
        super(ViewCollection, self).__init__(context, View, resource_path, parent_list)

    def add(self, view_creation_information):
        """

        :type view_creation_information: office365.sharepoint.view_create_information.ViewCreationInformation
        """
        view = View(self.context, None, self._parent)
        self.add_child(view)
        qry = ServiceOperationQuery(self, "Add", None, view_creation_information, "parameters", view)
        self.context.add_query(qry)
        return view

    def get_by_title(self, view_title):
        """Gets the list view with the specified title.

        :type view_title: str
        """
        return View(self.context,
                    ResourcePathServiceOperation("GetByTitle", [view_title], self.resource_path), self._parent)

    def get_by_id(self, view_id):
        """Gets the list view with the specified ID.

        :type view_id: str
        """
        return View(self.context,
                    ResourcePathServiceOperation("GetById", [view_id], self.resource_path), self._parent)

    @property
    def parent_list(self):
        return self._parent



from office365.runtime.client_value import ClientValue


class Location(ClientValue):
    """Represents location information of an event."""

    def __init__(self, displayName=None):
        """

        :param str displayName: The name associated with the location.
        """
        super(Location, self).__init__()
        self.displayName = displayName



from office365.directory.identities.userflows.b2x_identity_user_flow import B2XIdentityUserFlow
from office365.directory.identities.identity_provider_base import IdentityProviderBase
from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.runtime.resource_path import ResourcePath


class IdentityContainer(Entity):

    @property
    def identity_providers(self):
        return self.properties.get('identityProviders',
                                   EntityCollection(self.context, IdentityProviderBase,
                                                    ResourcePath("identityProviders", self.resource_path)))

    @property
    def b2x_user_flows(self):
        return self.properties.get('b2xUserFlows',
                                   EntityCollection(self.context, B2XIdentityUserFlow,
                                                    ResourcePath("b2xUserFlows", self.resource_path)))



from office365.runtime.client_value import ClientValue


class NumberColumn(ClientValue):
    """The numberColumn on a columnDefinition resource indicates that the column's values are numbers."""
    pass



from office365.onenote.entity_schema_object_model import OnenoteEntitySchemaObjectModel
from office365.onenote.notebooks.notebook import Notebook
from office365.onenote.pages.page_links import PageLinks
from office365.onenote.sections.section import OnenoteSection
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class OnenotePage(OnenoteEntitySchemaObjectModel):
    """A page in a OneNote notebook."""

    def get_content(self):
        """Download the page's HTML content. """
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "content", None, None, None, result)

        def _construct_query(request):
            """
            :type request: office365.runtime.http.request_options.RequestOptions
            """
            request.method = HttpMethod.Get

        self.context.before_execute(_construct_query)
        self.context.add_query(qry)
        return result

    @property
    def content_url(self):
        """The URL for the page's HTML content. Read-only.

        :rtype: str or None
        """
        return self.properties.get("contentUrl", None)

    @property
    def links(self):
        """Links for opening the page. The oneNoteClientURL link opens the page in the OneNote native client
        if it 's installed. The oneNoteWebUrl link opens the page in OneNote on the web. Read-only.

        """
        return self.properties.get("links", PageLinks())

    @property
    def user_tags(self):
        """Links for opening the page. The oneNoteClientURL link opens the page in the OneNote native client
        if it 's installed. The oneNoteWebUrl link opens the page in OneNote on the web. Read-only.

        """
        return self.properties.get("userTags", ClientValueCollection(str))

    @property
    def parent_notebook(self):
        """The notebook that contains the page. Read-only.

        :rtype: Notebook
        """
        return self.get_property('parentNotebook',
                                 Notebook(self.context, ResourcePath("parentNotebook", self.resource_path)))

    @property
    def parent_section(self):
        """The section that contains the page. Read-only.

        :rtype: OnenoteSection
        """
        return self.get_property('parentSection',
                                 OnenoteSection(self.context, ResourcePath("parentSection", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "userTags": self.user_tags,
                "parentSection": self.parent_section,
                "parentNotebook": self.parent_notebook
            }
            default_value = property_mapping.get(name, None)
        return super(OnenotePage, self).get_property(name, default_value)



from office365.runtime.queries.client_query import ClientQuery


class UpdateEntityQuery(ClientQuery):
    def __init__(self, entity_to_update):
        """
        Update entity query

        :type entity_to_update: office365.runtime.client_object.ClientObject
        """
        super(UpdateEntityQuery, self).__init__(entity_to_update.context,
                                                entity_to_update,
                                                entity_to_update,
                                                None,
                                                None)



from office365.runtime.client_value import ClientValue


class OnenoteOperationError(ClientValue):
    """An error from a failed OneNote operation."""

    def __init__(self, message=None, code=None):
        """
        :param str message: The error message.
        :param str code: The error code.
        """
        super(OnenoteOperationError, self).__init__()
        self.message = message
        self.code = code



class ClientResult(object):
    """Client result"""

    def __init__(self, context, default_value=None):
        """

        :type context: office365.runtime.client_runtime_context.ClientRuntimeContext
        :type default_value: any
        """
        self.context = context
        self.value = default_value

    def build_request(self):
        return self.context.build_request()

    def execute_query(self,**kwargs):
        self.context.execute_query(**kwargs)
        return self

    def execute_query_retry(self, max_retry=5, timeout_secs=5, success_callback=None, failure_callback=None,**kwargs):
        self.context.execute_query_retry(max_retry=max_retry,
                                         timeout_secs=timeout_secs,
                                         success_callback=success_callback,
                                         failure_callback=failure_callback,
                                         **kwargs)
        return self




from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection
from office365.search.search_hits_container import SearchHitsContainer


class SearchResponse(ClientValue):
    """Represents results from a search query, and the terms used for the query."""

    def __init__(self, search_terms=None, hits_containers=None):
        """

        """
        super(SearchResponse, self).__init__()
        self.searchTerms = search_terms
        if hits_containers is None:
            hits_containers = ClientValueCollection(SearchHitsContainer)
        self.hitsContainers = hits_containers



from office365.runtime.odata.odata_reader import ODataReader


class ODataV3Reader(ODataReader):
    """OData v3 reader"""

    def __init__(self, metadata_path):
        super(ODataV3Reader, self).__init__(metadata_path, {
            'xmlns': 'http://schemas.microsoft.com/ado/2009/11/edm',
            'edmx': 'http://schemas.microsoft.com/ado/2007/06/edmx',
            'm': 'http://schemas.microsoft.com/ado/2007/08/dataservices/metadata'
        })




from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity import BaseEntity


class Principal(BaseEntity):
    """Represents a user or group that can be assigned permissions to control security."""

    @property
    def id(self):
        """Gets a value that specifies the member identifier for the user or group.

        :rtype: int or None
        """
        return self.properties.get('Id', None)

    @property
    def title(self):
        """Gets a value that specifies the name of the principal.

        :rtype: str or None
        """
        return self.properties.get('Title', None)

    @title.setter
    def title(self, value):
        self.set_property('Title', value)

    @property
    def login_name(self):
        """Gets the login name of the principal.

        :rtype: str or None
        """
        return self.properties.get('LoginName', None)

    @property
    def user_principal_name(self):
        """Gets the UPN of the principal.

        :rtype: str or None
        """
        return self.properties.get('UserPrincipalName', None)

    @property
    def is_hidden_in_ui(self):
        """Gets the login name of the principal.

        :rtype: bool or None
        """
        return self.properties.get('IsHiddenInUI', None)

    @property
    def principal_type(self):
        """Gets the login name of the principal.

        :rtype: int or None
        """
        return self.properties.get('PrincipalType', None)

    def set_property(self, name, value, persist_changes=True):
        super(Principal, self).set_property(name, value, persist_changes)
        # fallback: create a new resource path
        if self._resource_path is None:
            if name == "Id":
                self._resource_path = ResourcePathServiceOperation(
                    "GetById", [value], self._parent_collection.resource_path)
            elif name == "LoginName":
                self._resource_path = ResourcePathServiceOperation(
                    "GetByName", [value], self._parent_collection.resource_path)
        return self



from office365.runtime.queries.client_query import ClientQuery


class CreateEntityQuery(ClientQuery):
    def __init__(self, parent_entity, parameters, entity_to_create):
        """
        Create entity query

        :type entity_to_create: office365.runtime.client_object.ClientObject
        :type parameters: office365.runtime.client_object.ClientObject or office365.runtime.client_value.ClientValue
            or dict
        :type parent_entity: office365.runtime.client_object.ClientObject
        """
        super(CreateEntityQuery, self).__init__(parent_entity.context,
                                                parent_entity,
                                                parameters,
                                                None,
                                                entity_to_create)



from office365.runtime.client_object import ClientObject
from office365.runtime.queries.delete_entity_query import DeleteEntityQuery
from office365.runtime.queries.update_entity_query import UpdateEntityQuery


class BaseEntity(ClientObject):

    def __init__(self, context, resource_path=None, namespace="SP", parent_collection=None):
        """
        SharePoint specific entity

        :param office365.sharepoint.client_context.ClientContext context: SharePoint context
        :param ResourcePath resource_path: Resource Path
        :param str namespace:  default namespace
        """
        super(BaseEntity, self).__init__(context, resource_path, parent_collection, namespace)

    def with_credentials(self, credentials):
        """
        :type credentials:  UserCredential or ClientCredential
        """
        self.context.with_credentials(credentials)
        return self

    def delete_object(self):
        """The recommended way to delete a SharePoint entity"""
        qry = DeleteEntityQuery(self)
        self.context.add_query(qry)
        self.remove_from_parent_collection()
        return self

    def update(self, *args):
        """The recommended way to update a SharePoint entity"""
        qry = UpdateEntityQuery(self)
        self.context.add_query(qry)
        return self

    @property
    def context(self):
        """
        :rtype: office365.sharepoint.client_context.ClientContext
        """
        return self._context

    @property
    def entity_type_name(self):
        if self._entity_type_name is None:
            self._entity_type_name = ".".join([self._namespace, type(self).__name__])
        return self._entity_type_name



from office365.runtime.client_value import ClientValue


class Shared(ClientValue):

    def __init__(self, owner=None):
        """
        The Shared resource indicates a DriveItem has been shared with others. The resource includes information
        about how the item is shared.

        :type owner: office365.directory.identitySet.IdentitySet
        """
        super(Shared, self).__init__()
        self.owner = owner



import uuid

from office365.runtime.queries.client_query import ClientQuery
from office365.runtime.queries.read_entity_query import ReadEntityQuery


def create_boundary(prefix, compact=False):
    """Creates a string that can be used as a multipart request boundary.

    :param bool compact:
    :param str prefix: String to use as the start of the boundary string
    """
    if compact:
        return prefix + str(uuid.uuid4())[:8]
    else:
        return prefix + str(uuid.uuid4())


class BatchQuery(ClientQuery):
    """Client query collection"""

    def __init__(self, context, queries=None):
        """

        :type context: office365.runtime.client_runtime_context.ClientRuntimeContext
        :type queries: list[ClientQuery]
        """
        super(BatchQuery, self).__init__(context)
        self._current_boundary = create_boundary("batch_")
        if queries is None:
            queries = []
        self._queries = queries

    def add(self, query):
        """

        :type query: ClientQuery
        """
        self._queries.append(query)

    @property
    def ordered_queries(self):
        return self.change_sets + self.get_queries

    @property
    def current_boundary(self):
        return self._current_boundary

    @property
    def change_sets(self):
        return [qry for qry in self._queries if not isinstance(qry, ReadEntityQuery)]

    @property
    def queries(self):
        """
        :rtype: list[ClientQuery]
        """
        return self._queries

    @property
    def get_queries(self):
        return [qry for qry in self._queries if isinstance(qry, ReadEntityQuery)]

    @property
    def has_change_sets(self):
        return len(self.change_sets) > 0



from office365.directory.identities.identity import Identity
from office365.runtime.client_value import ClientValue


class IdentitySet(ClientValue):
    """The IdentitySet resource is a keyed collection of identity resources. It is used to represent a set of
    identities associated with various events for an item, such as created by or last modified by. """

    def __init__(self):
        super(IdentitySet, self).__init__()
        self.application = Identity()
        self.device = Identity()
        self.user = Identity()



from office365.runtime.client_value import ClientValue


class ChangeNotificationCollection(ClientValue):
    """Represents a collection of resource change notifications sent to the subscriber."""

    def __init__(self, validation_tokens=None, value=None):
        super(ChangeNotificationCollection, self).__init__()
        self.validationTokens = validation_tokens
        self.value = value



from office365.sharepoint.base_entity import BaseEntity


class WebTemplate(BaseEntity):
    """Specifies a site definition or a site template that is used to instantiate a site."""

    @property
    def name(self):
        """Gets a value that specifies the display name of the list template."""
        return self.properties.get('Name', None)

    @property
    def description(self):
        """Gets a value that specifies the description of the list template."""
        return self.properties.get('Description', None)



from office365.outlook.mail.attachment import Attachment


class FileAttachment(Attachment):
    """A file (such as a text file or Word document) attached to a user event, message, or post."""

    @property
    def content_id(self):
        """
        The ID of the attachment in the Exchange store.

        :rtype: str or None
        """
        return self.properties.get("contentId", None)

    @property
    def content_location(self):
        """
        :rtype: str or None
        """
        return self.properties.get("content_location", None)

    @property
    def content_bytes(self):
        """
        The base64-encoded contents of the file.

        :rtype: str or None
        """
        return self.properties.get("contentBytes", None)

    @content_bytes.setter
    def content_bytes(self, value):
        """
        Sets the base64-encoded contents of the file.
        """
        self.set_property("contentBytes", value)



from office365.runtime.client_value import ClientValue


class TextColumn(ClientValue):
    """The textColumn on a columnDefinition resource indicates that the column's values are text."""

    def __init__(self, max_length=None):
        super(TextColumn, self).__init__()
        self.maxLength = max_length



from office365.runtime.client_runtime_context import ClientRuntimeContext
from office365.runtime.odata.odata_request import ODataRequest
from office365.runtime.odata.v4.json_format import V4JsonFormat
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.taxonomy.term_store import TermStore


class TaxonomyService(ClientRuntimeContext):
    """Wraps all of the associated TermStore objects for an Site object."""

    def __init__(self, context):
        """
        :type  context: office365.sharepoint.client_context.ClientContext
        """
        super(TaxonomyService, self).__init__()
        self._auth_context = context.authentication_context
        self._pendingRequest = ODataRequest(self, V4JsonFormat("minimal"))
        self._service_root_url = "{0}/v2.1/".format(context.service_root_url())

    def authenticate_request(self, request):
        self._auth_context.authenticate_request(request)

    def pending_request(self):
        return self._pendingRequest

    def service_root_url(self):
        return self._service_root_url

    @property
    def term_store(self):
        return TermStore(self, ResourcePath("termStore", None))



from office365.runtime.client_value import ClientValue


class FieldGeolocationValue(ClientValue):

    def __init__(self, latitude, longitude, altitude=None):
        """
        :param float latitude: Specifies the latitude value for Geolocation field.
        :param float longitude: Specifies the longitude value for Geolocation field.
        :param float altitude: Specifies the altitude value for Geolocation field. It is a user defined value

        """
        super(FieldGeolocationValue, self).__init__()
        self.Latitude = latitude
        self.Longitude = longitude
        self.Altitude = altitude

    @property
    def entity_type_name(self):
        return "SP.FieldGeolocationValue"



from office365.entity import Entity


class PlannerTaskDetails(Entity):
    """
    The plannerTaskDetails resource represents the additional information about a task.
    Each task object has a details object.
    """
    pass



from office365.entity import Entity


class TeamsTemplate(Entity):
    """A team template is a blueprint for creating a team in Microsoft Teams.
    A template specifies the structure, settings, and even content that should be provisioned in a new team created
    using the template. Microsoft provides a suite of base templates
    and customers can save their own custom templates."""
    pass



from office365.runtime.client_value import ClientValue
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.taxonomy.taxonomy_item import TaxonomyItem, TaxonomyItemCollection
from office365.sharepoint.taxonomy.term import Term


class TermSet(TaxonomyItem):

    @property
    def terms(self):
        return self.properties.get("terms",
                                   TaxonomyItemCollection(self.context, Term,
                                                          ResourcePath("terms", self.resource_path)))


class TermSetInfo(ClientValue):
    pass



from office365.directory.directory_object_collection import DirectoryObjectCollection
from office365.directory.directory_object import DirectoryObject
from office365.runtime.resource_path import ResourcePath


class ServicePrincipal(DirectoryObject):
    """Represents an instance of an application in a directory."""

    def add_key(self, key_credential, password_credential, proof):
        pass

    def add_password(self):
        pass

    @property
    def app_display_name(self):
        """The collection of key credentials associated with the application. Not nullable.
        """
        return self.properties.get('appDisplayName', None)

    @property
    def service_principal_type(self):
        """
        Identifies whether the service principal represents an application, a managed identity, or a legacy application.
        This is set by Azure AD internally. The servicePrincipalType property can be set to three different values:
            Application - A service principal that represents an application or service.
            The appId property identifies the associated app registration, and matches the appId of an application,
            possibly from a different tenant. If the associated app registration is missing, tokens are not issued
            for the service principal.

            ManagedIdentity - A service principal that represents a managed identity. Service principals
            representing managed identities can be granted access and permissions, but cannot be updated
            or modified directly.

            Legacy - A service principal that represents an app created before app registrations,
            or through legacy experiences. Legacy service principal can have credentials, service principal names,
            reply URLs, and other properties which are editable by an authorized user,
            but does not have an associated app registration. The appId value does not associate
            the service principal with an app registration.
            The service principal can only be used in the tenant where it was created.

        :rtype: str or None
        """
        return self.properties.get('servicePrincipalType', None)

    @property
    def owners(self):
        """Directory objects that are owners of this servicePrincipal.
        The owners are a set of non-admin users or servicePrincipals who are allowed to modify this object.

        :rtype: DirectoryObjectCollection
        """
        return self.get_property('owners',
                                 DirectoryObjectCollection(self.context, ResourcePath("owners", self.resource_path)))



from office365.entity import Entity
from office365.outlook.calendar.email_address import EmailAddress


class CalendarPermission(Entity):
    """
    The permissions of a user with whom the calendar has been shared or delegated in an Outlook client.

    Get, update, and delete of calendar permissions is supported on behalf of only the calendar owner.

    Getting the calendar permissions of a calendar on behalf of a sharee or delegate returns
    an empty calendar permissions collection.

    Once a sharee or delegate has been set up for a calendar, you can update only the role property to change
    the permissions of a sharee or delegate. You cannot update the allowedRoles, emailAddress, isInsideOrganization,
    or isRemovable property. To change these properties, you should delete the corresponding calendarPermission
    object and create another sharee or delegate in an Outlook client.

    """

    @property
    def email_address(self):
        """
        Represents a sharee or delegate who has access to the calendar.
        For the "My Organization" sharee, the address property is null. Read-only.
        """
        return self.properties.get("emailAddress", EmailAddress())



class ViewType:
    """Specifies the type of the list view."""
    def __init__(self):
        pass

    NoneType = 0
    Html = 1
    Grid = 2048
    Calendar = 524288
    Recurrence = 8193
    Chart = 131072
    Gantt = 67108864



from office365.runtime.client_value import ClientValue


class ListCreationInformation(ClientValue):
    """Represents metadata about list creation."""

    def __init__(self, title=None, description=None, base_template=None, allow_content_types=False):
        """

        :type base_template: int or None
        :type allow_content_types: bool
        :type description: str or None
        :type title: str
        """
        super(ListCreationInformation, self).__init__()
        self.Title = title
        self.Description = description
        self.BaseTemplate = base_template
        self.AllowContentTypes = allow_content_types

    @property
    def entity_type_name(self):
        return "SP.List"



from office365.directory.directory_object_collection import DirectoryObjectCollection
from office365.directory.directory_object import DirectoryObject
from office365.directory.extensions.extension_property import ExtensionProperty
from office365.directory.key_credential import KeyCredential
from office365.directory.password_credential import PasswordCredential
from office365.entity_collection import EntityCollection
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class Application(DirectoryObject):
    """
    Represents an application. Any application that outsources authentication to Azure Active Directory (Azure AD)
    must be registered in a directory. Application registration involves telling Azure AD about your application,
    including the URL where it's located, the URL to send replies after authentication,
    the URI to identify your application, and more. For more information, see Basics of Registering
    an Application in Azure AD
    """

    def add_password(self, display_name):
        """Adds a strong password to an application.
        :param str display_name: App display name
        """
        params = PasswordCredential(displayName=display_name)
        result = ClientResult(self.context, params)
        qry = ServiceOperationQuery(self, "addPassword", None, params, None, result)
        self.context.add_query(qry)
        return result

    def remove_password(self, key_id):
        """Remove a password from an application."""
        qry = ServiceOperationQuery(self, "removePassword", None, {"keyId": key_id})
        self.context.add_query(qry)
        return self

    def delete_object(self, permanent_delete=False):
        """
        :param permanent_delete: Permanently deletes the application from directory
        :type permanent_delete: bool

        """
        super(Application, self).delete_object()
        if permanent_delete:
            deleted_item = self.context.directory.deleted_applications[self.id]
            deleted_item.delete_object()
        return self

    def set_verified_publisher(self, verified_publisher_id):
        """Set the verifiedPublisher on an application.
        For more information, including prerequisites to setting a verified publisher, see Publisher verification.

        :param str verified_publisher_id: The Microsoft Partner Network ID (MPNID) of the verified publisher
        to be set on the application, from the publisher's Partner Center account.
        """
        qry = ServiceOperationQuery(self, "setVerifiedPublisher", None, {"verifiedPublisherId": verified_publisher_id})
        self.context.add_query(qry)
        return self

    def unset_verified_publisher(self):
        """Unset the verifiedPublisher previously set on an application, removing all verified publisher properties.
        For more information, see Publisher verification.
        """
        qry = ServiceOperationQuery(self, "unsetVerifiedPublisher")
        self.context.add_query(qry)
        return self

    def add_key(self, key_credential, password_credential, proof):
        """
        Add a key credential to an application. This method, along with removeKey can be used by an application
        to automate rolling its expiring keys.

        :param KeyCredential key_credential: The new application key credential to add.
            The type, usage and key are required properties for this usage. Supported key types are:
                AsymmetricX509Cert: The usage must be Verify.
                X509CertAndPassword: The usage must be Sign
        :param PasswordCredential password_credential: Only secretText is required to be set which should contain the password
             for the key. This property is required only for keys of type X509CertAndPassword. Set it to null otherwise.
        :param str proof: A self-signed JWT token used as a proof of possession of the existing keys
        """
        payload = {
            "keyCredential": key_credential,
            "passwordCredential": password_credential,
            "proof": proof,
        }
        return_type = ClientResult(self.context, KeyCredential())
        qry = ServiceOperationQuery(self, "addKey", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def remove_key(self, keyId, proof):
        """
        Remove a key credential from an application.
        This method along with addKey can be used by an application to automate rolling its expiring keys.

        :param str keyId: The unique identifier for the password.
        :param str proof: A self-signed JWT token used as a proof of possession of the existing keys.
             This JWT token must be signed using the private key of one of the application's existing
             valid certificates. The token should contain the following claims:
                 aud - Audience needs to be 00000002-0000-0000-c000-000000000000.
                 iss - Issuer needs to be the id of the application that is making the call.
                 nbf - Not before time.
                 exp - Expiration time should be "nbf" + 10 mins.
        """
        qry = ServiceOperationQuery(self, "removeKey", None, {"keyId": keyId, "proof": proof})
        self.context.add_query(qry)
        return self

    @property
    def key_credentials(self):
        """The collection of key credentials associated with the application. Not nullable.
        """
        return self.properties.get('keyCredentials', ClientValueCollection(KeyCredential))

    @property
    def display_name(self):
        """
        The display name for the application.
        Supports $filter (eq, ne, NOT, ge, le, in, startsWith), $search, and $orderBy.

        :rtype: str or None
        """
        return self.properties.get('displayName', None)

    @property
    def identifier_uris(self):
        """
        The URIs that identify the application within its Azure AD tenant, or within a verified custom domain
        if the application is multi-tenant. For more information see Application Objects and Service Principal Objects.
        The any operator is required for filter expressions on multi-valued properties.
        """
        return self.properties.get('identifierUris', ClientValueCollection(str))

    @property
    def signin_audience(self):
        """
        Specifies the Microsoft accounts that are supported for the current application.
        Supported values are: AzureADMyOrg, AzureADMultipleOrgs, AzureADandPersonalMicrosoftAccount,
        PersonalMicrosoftAccount

        :rtype: str or None
        """
        return self.properties.get('signInAudience', None)

    @property
    def owners(self):
        """Directory objects that are owners of the application. Read-only.

        :rtype: DirectoryObjectCollection
        """
        return self.get_property('owners',
                                 DirectoryObjectCollection(self.context, ResourcePath("owners", self.resource_path)))

    @property
    def extension_properties(self):
        """List extension properties on an application object.

        :rtype: EntityCollection
        """
        return self.get_property('extensionProperties',
                                 EntityCollection(self.context, ExtensionProperty,
                                                  ResourcePath("extensionProperties", self.resource_path)))



from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.forms.form import Form


class FormCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None):
        super(FormCollection, self).__init__(context, Form, resource_path)

    def get_by_id(self, _id):
        """Gets the form with the specified ID."""
        return Form(self.context, ResourcePathServiceOperation("GetById", [_id], self.resource_path))

    def get_by_page_type(self, form_type):
        return Form(self.context, ResourcePathServiceOperation("GetByPageType", [form_type], self.resource_path))



from office365.directory.identities.identity_set import IdentitySet
from office365.entity import Entity
from office365.onedrive.listitems.item_reference import ItemReference


class BaseItem(Entity):
    """The baseItem resource is an abstract resource that contains a common set of properties shared among several
    other resources types """

    @property
    def etag(self):
        """ETag for the item."""
        return self.properties.get('eTag', None)

    @property
    def created_by(self):
        """Identity of the user, device, or application which created the item."""
        return self.properties.get('createdBy', IdentitySet())

    @property
    def last_modified_by(self):
        """Identity of the user, device, and application which last modified the item."""
        return self.properties.get('lastModifiedBy', IdentitySet())

    @property
    def created_datetime(self):
        """Gets date and time of item creation."""
        return self.properties.get('createdDateTime', None)

    @property
    def last_modified_datetime(self):
        """Gets date and time the item was last modified."""
        return self.properties.get('lastModifiedDateTime', None)

    @property
    def name(self):
        """Gets the name of the item."""
        return self.properties.get('name', None)

    @name.setter
    def name(self, value):
        """
        Sets the name of the item.

        :type value: str
        """
        self.set_property('name', value)

    @property
    def description(self):
        """
        Provides a user-visible description of the item.

        :rtype: str or None
        """
        return self.properties.get('description', None)

    @description.setter
    def description(self, value):
        self.set_property('description', value)

    @property
    def web_url(self):
        """
        URL that displays the resource in the browser.

        :rtype: str or None
        """
        return self.properties.get('webUrl', None)

    @property
    def parent_reference(self):
        """Parent information, if the item has a parent."""
        return self.properties.get('parentReference', ItemReference())

    @parent_reference.setter
    def parent_reference(self, value):
        self.set_property('parentReference', value, False)

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "createdBy": self.created_by,
                "lastModifiedBy": self.last_modified_by,
                "parentReference": self.parent_reference
            }
            default_value = property_mapping.get(name, None)
        return super(BaseItem, self).get_property(name, default_value)




from office365.sharepoint.base_entity import BaseEntity


class LikedByInformation(BaseEntity):

    @property
    def like_count(self):
        return self.properties.get("LikeCount", None)

    def set_property(self, name, value, persist_changes=True):
        super(LikedByInformation, self).set_property(name, value, persist_changes)
        return self





from office365.runtime.client_value import ClientValue


class DefaultColumnValue(ClientValue):
    """The defaultColumnValue on a columnDefinition resource specifies the default value for this column.
    The default value can either be specified directly or as a formula."""

    def __init__(self, formula=None, value=None):
        super(DefaultColumnValue, self).__init__()
        self.formula = formula
        self.value = value



from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.permissions.role_definition import RoleDefinition


class RoleDefinitionCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None):
        super(RoleDefinitionCollection, self).__init__(context, RoleDefinition, resource_path)

    def get_by_type(self, role_type):
        """Returns role definition of the specified type from the collection.

        :param int role_type: Specifies the role type. Role type MUST NOT be None.
        """
        return RoleDefinition(self.context,
                              ResourcePathServiceOperation("GetByType", [role_type], self.resource_path))



from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.changes.change_collection import ChangeCollection
from office365.sharepoint.changes.change_query import ChangeQuery
from office365.sharepoint.contenttypes.content_type_collection import ContentTypeCollection
from office365.sharepoint.customactions.custom_action_element import CustomActionElement
from office365.sharepoint.eventreceivers.event_receiver_definition import EventReceiverDefinitionCollection
from office365.sharepoint.fields.field_collection import FieldCollection
from office365.sharepoint.fields.related_field_collection import RelatedFieldCollection
from office365.sharepoint.files.checkedOutFileCollection import CheckedOutFileCollection
from office365.sharepoint.files.file import File
from office365.sharepoint.flows.flow_synchronization_result import FlowSynchronizationResult
from office365.sharepoint.folders.folder import Folder
from office365.sharepoint.forms.form_collection import FormCollection
from office365.sharepoint.listitems.caml.caml_query import CamlQuery
from office365.sharepoint.listitems.creation_information_using_path import ListItemCreationInformationUsingPath
from office365.sharepoint.listitems.form_update_value import ListItemFormUpdateValue
from office365.sharepoint.listitems.listitem import ListItem
from office365.sharepoint.listitems.listItem_collection import ListItemCollection
from office365.sharepoint.pages.wiki_page_creation_information import WikiPageCreationInformation
from office365.sharepoint.permissions.securable_object import SecurableObject
from office365.sharepoint.usercustomactions.user_custom_action_collection import UserCustomActionCollection
from office365.sharepoint.views.view import View
from office365.sharepoint.views.view_collection import ViewCollection
from office365.sharepoint.webhooks.subscription_collection import SubscriptionCollection
from office365.sharepoint.utilities.utility import Utility


class List(SecurableObject):
    """Represents a list on a SharePoint Web site."""

    def __init__(self, context, resource_path=None):
        super(List, self).__init__(context, resource_path)

    def get_business_app_operation_status(self):
        return_type = FlowSynchronizationResult(self.context)
        qry = ServiceOperationQuery(self, "GetBusinessAppOperationStatus", None, None, None, return_type)
        self.context.add_query(qry)
        return return_type

    def sync_flow_callback_url(self, flow_id):
        """
        :param str flow_id:
        """
        return_type = FlowSynchronizationResult(self.context)
        qry = ServiceOperationQuery(self, "SyncFlowCallbackUrl", None, {"flowId": flow_id}, None, return_type)
        self.context.add_query(qry)
        return return_type

    def sync_flow_instance(self, flow_id):
        """
        :param str flow_id:
        """
        return_type = FlowSynchronizationResult(self.context)
        qry = ServiceOperationQuery(self, "SyncFlowInstance", None, {"flowId": flow_id}, None, return_type)
        self.context.add_query(qry)
        return return_type

    def sync_flow_instances(self, retrieve_group_flows):
        """
        :param bool retrieve_group_flows:
        """
        return_type = FlowSynchronizationResult(self.context)
        payload = {"retrieveGroupFlows": retrieve_group_flows}
        qry = ServiceOperationQuery(self, "SyncFlowInstances", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def sync_flow_templates(self, category):
        """
        :param str category:
        """
        return_type = FlowSynchronizationResult(self.context)
        payload = {"category": category}
        qry = ServiceOperationQuery(self, "SyncFlowTemplates", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def create_document_with_default_name(self, folder_path, extension):
        """
        :param str folder_path:
        :param str extension:
        """
        return_type = FlowSynchronizationResult(self.context)
        payload = {
            "folderPath": folder_path,
            "extension": extension
        }
        qry = ServiceOperationQuery(self, "CreateDocumentWithDefaultName", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def recycle(self):
        """Moves the list to the Recycle Bin and returns the identifier of the new Recycle Bin item."""
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "Recycle", None, None, None, result)
        self.context.add_query(qry)
        return result

    def render_list_data(self, view_xml):
        """

        :param str view_xml: View xml
        """
        result = ClientResult(self.context)
        payload = {
            "viewXml": view_xml
        }
        qry = ServiceOperationQuery(self, "RenderListData", None, payload, None, result)
        self.context.add_query(qry)
        return result

    @staticmethod
    def get_list_data_as_stream(context, list_full_url, parameters=None):
        """

        :param office365.sharepoint.client_context.ClientContext context:
        :param str list_full_url:
        :param RenderListDataParameters parameters:
        """
        result = ClientResult(context)
        payload = {
            "listFullUrl": list_full_url,
            "parameters": parameters,
        }
        target_list = context.web.get_list(list_full_url)
        qry = ServiceOperationQuery(target_list, "GetListDataAsStream", None, payload, None, result)
        context.add_query(qry)
        return result

    def bulk_validate_update_list_items(self, item_ids, form_values, new_document_update=True,
                                        checkin_comment=None, folder_path=None):
        """

        :param list[int] item_ids:
        :param dict form_values:
        :param bool new_document_update:
        :param str checkin_comment:
        :param str folder_path:
        """
        result = ClientValueCollection(ListItemFormUpdateValue)
        params = {
            "itemIds": item_ids,
            "formValues": ClientValueCollection(ListItemFormUpdateValue, form_values),
            "bNewDocumentUpdate": new_document_update,
            "checkInComment": checkin_comment,
            "folderPath": folder_path
        }
        qry = ServiceOperationQuery(self, "BulkValidateUpdateListItems", None, params, None, result)
        self.context.add_query(qry)
        return result

    def get_lookup_field_choices(self, targetFieldName, pagingInfo=None):
        """

        """
        result = ClientResult(self.context)
        params = {
            "targetFieldName": targetFieldName,
            "pagingInfo": pagingInfo
        }
        qry = ServiceOperationQuery(self, "GetLookupFieldChoices", params, None, None, result)
        self.context.add_query(qry)
        return result

    def get_list_item_changes_since_token(self, query):
        """

        :type query: office365.sharepoint.changes.change_log_item_query.ChangeLogItemQuery
        """
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "getListItemChangesSinceToken", None, query, "query", result)
        self.context.add_query(qry)
        return result

    def save_as_template(self, file_name, name, description, save_data):
        """
        Saves the list as a template in the list template gallery and includes the option of saving with or
        without the data that is contained in the current list.

        :param bool save_data: true to save the data of the original list along with the list template; otherwise, false.
        :param str description: A string that contains the description for the list template.
        :param str name: A string that contains the title for the list template.
        :param str file_name: A string that contains the file name for the list template with an .stp extension.
        :return:
        """
        payload = {
            "strFileName": file_name,
            "strName": name,
            "strDescription": description,
            "bSaveData": save_data
        }
        qry = ServiceOperationQuery(self, "saveAsTemplate", None, payload, None, None)
        self.context.add_query(qry)
        return self

    def get_item_by_unique_id(self, unique_id):
        """
        Returns the list item with the specified ID.

        :param str unique_id:"""
        item = ListItem(self.context, ResourcePathServiceOperation("getItemByUniqueId", [unique_id], self.resource_path))
        return item

    def get_web_dav_url(self, source_url):
        """
        Gets the trusted URL for opening the folder in Explorer view.

        :param str source_url: The URL of the current folder the user is in.
        :return: ClientResult
        """

        result = ClientResult(self.context)
        payload = {
            "sourceUrl": source_url
        }
        qry = ServiceOperationQuery(self, "getWebDavUrl", None, payload, None, result)
        self.context.add_query(qry)
        return result

    def get_items(self, caml_query=None):
        """Returns a collection of items from the list based on the specified query.
        :type caml_query: CamlQuery
        """
        if not caml_query:
            caml_query = CamlQuery.create_all_items_query()
        items = ListItemCollection(self.context, ResourcePath("items", self.resource_path))
        qry = ServiceOperationQuery(self, "GetItems", None, caml_query, "query", items)
        self.context.add_query(qry)
        return items

    def add_item(self, list_item_creation_information):
        """The recommended way to add a list item is to send a POST request to the ListItemCollection resource endpoint,
         as shown in ListItemCollection request examples.

         :type list_item_creation_information: ListItemCreationInformation or dict"""
        item = ListItem(self.context, None, self)
        if isinstance(list_item_creation_information, dict):
            for k, v in list_item_creation_information.items():
                item.set_property(k, v, True)
            self.items.add_child(item)
            item.ensure_type_name(self)
            qry = ServiceOperationQuery(self, "items", None, item, None, item)
            self.context.add_query(qry)
        else:
            def _resolve_folder_url():
                list_item_creation_information.FolderUrl = self.context.base_url + self.root_folder.serverRelativeUrl
                add_item_qry = ServiceOperationQuery(
                    self,
                    "addItem",
                    None,
                    list_item_creation_information,
                    "parameters",
                    item
                )
                self.context.add_query(add_item_qry)

            self.root_folder.ensure_property("ServerRelativeUrl", _resolve_folder_url)
        return item

    def create_wiki_page(self, page_name, page_content):
        """
        :param str page_name:
        :param str page_content:
        """
        result = ClientResult(self.context, File(self.context))

        def _list_loaded():
            page_url = self.root_folder.serverRelativeUrl + "/" + page_name
            wiki_props = WikiPageCreationInformation(page_url, page_content)
            result.value = Utility.create_wiki_page_in_context_web(self.context, wiki_props)
        self.ensure_property("RootFolder", _list_loaded)

        return result

    def add_item_using_path(self, leaf_name, object_type, folder_url):
        """
        :type leaf_name: str
        :type object_type: int
        :type folder_url: str
        """
        from office365.sharepoint.types.resource_path import ResourcePath as SPResPath
        parameters = ListItemCreationInformationUsingPath(leaf_name, object_type, folder_path=SPResPath(folder_url))
        item = ListItem(self.context)
        qry = ServiceOperationQuery(self, "AddItemUsingPath", None, parameters, "parameters", item)
        self.context.add_query(qry)
        return item

    def add_validate_update_item(self):
        pass

    def get_item_by_id(self, item_id):
        """Returns the list item with the specified list item identifier.

        :type item_id: int
        """
        return ListItem(self.context,
                        ResourcePathServiceOperation("getItemById", [item_id], self.resource_path))

    def get_view(self, view_id):
        """Returns the list view with the specified view identifier.

        :type view_id: str
        """
        view = View(self.context, ResourcePathServiceOperation("getView", [view_id], self.resource_path), self)
        return view

    def get_changes(self, query=None):
        """Returns the collection of changes from the change log that have occurred within the list,
           based on the specified query.

        :param office365.sharepoint.changeQuery.ChangeQuery query: Specifies which changes to return
        """
        if query is None:
            query = ChangeQuery(list_=True)
        changes = ChangeCollection(self.context)
        qry = ServiceOperationQuery(self, "getChanges", None, query, "query", changes)
        self.context.add_query(qry)
        return changes

    def get_checked_out_files(self):
        result = CheckedOutFileCollection(self.context)
        qry = ServiceOperationQuery(self, "GetCheckedOutFiles", None, None, None, result)
        self.context.add_query(qry)
        return result

    def get_related_fields(self):
        """Returns a collection of lookup fields that use this list as a data source and
            that have FieldLookup.IsRelationship set to true.
        """
        return RelatedFieldCollection(self.context,
                                      ResourcePathServiceOperation("getRelatedFields", [], self.resource_path))

    @property
    def id(self):
        """
        Gets a value that specifies the list identifier.
        :rtype: str
        """
        return self.properties.get("Id", None)

    @property
    def enable_folder_creation(self):
        """
        Specifies whether new list folders can be added to the list.

        :rtype: bool or None
        """
        return self.properties.get("EnableFolderCreation", None)

    @property
    def default_edit_form_url(self):
        """
        Gets a value that specifies the URL of the edit form to use for list items in the list.

        :rtype: str or None
        """
        return self.properties.get("DefaultEditFormUrl", None)

    @enable_folder_creation.setter
    def enable_folder_creation(self, value):
        self.set_property("EnableFolderCreation", value, True)

    @property
    def items(self):
        """Get list items"""
        return self.properties.get("Items",
                                   ListItemCollection(self.context, ResourcePath("items", self.resource_path)))

    @property
    def root_folder(self):
        """Get a root folder"""
        return self.properties.get("RootFolder",
                                   Folder(self.context, ResourcePath("RootFolder", self.resource_path)))

    @property
    def fields(self):
        """Gets a value that specifies the collection of all fields in the list."""
        return self.properties.get('Fields',
                                   FieldCollection(self.context, ResourcePath("Fields", self.resource_path), self))

    @property
    def subscriptions(self):
        """Gets one or more webhook subscriptions on a SharePoint list."""
        return self.properties.get('Subscriptions',
                                   SubscriptionCollection(self.context,
                                                          ResourcePath("Subscriptions", self.resource_path)))

    @property
    def views(self):
        """Gets a value that specifies the collection of all public views on the list and personal views
        of the current user on the list."""
        return self.properties.get('Views',
                                   ViewCollection(self.context, ResourcePath("views", self.resource_path), self))

    @property
    def default_view(self):
        """Gets or sets a value that specifies whether the list view is the default list view."""
        return self.properties.get('DefaultView',
                                   View(self.context, ResourcePath("DefaultView", self.resource_path), self))

    @property
    def content_types(self):
        """Gets the content types that are associated with the list."""
        return self.properties.get('ContentTypes',
                                   ContentTypeCollection(self.context,
                                                         ResourcePath("ContentTypes", self.resource_path), self))

    @property
    def user_custom_actions(self):
        """Gets the User Custom Actions that are associated with the list."""
        return self.properties.get('UserCustomActions',
                                   UserCustomActionCollection(self.context,
                                                              ResourcePath("UserCustomActions", self.resource_path)))

    @property
    def custom_action_elements(self):
        return self.properties.get('CustomActionElements',
                                   ClientValueCollection(CustomActionElement))

    @property
    def forms(self):
        """Gets a value that specifies the collection of all list forms in the list."""
        return self.properties.get('Forms',
                                   FormCollection(self.context, ResourcePath("forms", self.resource_path)))

    @property
    def parent_web(self):
        """Gets a value that specifies the web where list resides."""
        from office365.sharepoint.webs.web import Web
        return self.properties.get('ParentWeb',
                                   Web(self.context, ResourcePath("parentWeb", self.resource_path)))

    @property
    def event_receivers(self):
        """Get Event receivers"""
        return self.properties.get('EventReceivers',
                                   EventReceiverDefinitionCollection(self.context,
                                                                     ResourcePath("eventReceivers", self.resource_path),
                                                                     self))

    @property
    def item_count(self):
        """Gets a value that specifies the number of list items in the list.
        :rtype: int or None
        """
        return self.properties.get('ItemCount', None)

    @property
    def title(self):
        """Gets the displayed title for the list.
        :rtype: str or None
        """
        return self.properties.get('Title', None)

    @title.setter
    def title(self, val):
        """Sets the displayed title for the list."""
        self.set_property('Title', val)

    @property
    def description(self):
        """Gets the description for the list.
        :rtype: str or None
        """
        return self.properties.get('Description', None)

    @description.setter
    def description(self, val):
        """Sets the description for the list."""
        self.set_property('Description', val)

    @property
    def parent_web_path(self):
        return self.properties.get('ParentWebPath', None)

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "ContentTypes": self.content_types,
                "DefaultView": self.default_view,
                "ParentWeb": self.parent_web,
                "RootFolder": self.root_folder,
                "UserCustomActions": self.user_custom_actions
            }
            default_value = property_mapping.get(name, None)
        return super(List, self).get_property(name, default_value)

    def set_property(self, name, value, persist_changes=True):
        super(List, self).set_property(name, value, persist_changes)
        # fallback: create a new resource path
        if self._resource_path is None:
            if name == "Id":
                self._resource_path = ResourcePathServiceOperation(
                    "GetById", [value], self._parent_collection.resource_path)
        return self



from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity import BaseEntity


class RoleDefinition(BaseEntity):
    """Defines a single role definition, including a name, description, and set of rights."""

    @property
    def id(self):
        """Specifies the identifier of the role definition.
        Its value MUST be equal to or greater than 1073741824."""
        return self.properties.get('Id', None)

    @property
    def roleTypeKind(self):
        """Specifies the type of the role definition.
        Its value MUST be equal to or greater than 0. Its value MUST be equal to or less than 5."""
        return self.properties.get('RoleTypeKind', None)

    @property
    def name(self):
        """Gets a value that specifies the role definition name."""
        return self.properties.get('Name', None)

    @name.setter
    def name(self, value):
        """Sets a value that specifies the role definition name."""
        self.set_property('Name', value)

    @property
    def description(self):
        """Gets or sets a value that specifies the description of the role definition."""
        return self.properties.get('Description', None)

    @description.setter
    def description(self, value):
        """Gets or sets a value that specifies the description of the role definition."""
        self.set_property('Description', value)

    def set_property(self, name, value, persist_changes=True):
        if self.resource_path is None:
            if name == "Id":
                self._resource_path = ResourcePathServiceOperation(
                    "GetById", [value], self._parent_collection.resource_path)
        return super(RoleDefinition, self).set_property(name, value, persist_changes)



from office365.directory.directory_object import DirectoryObject
from office365.directory.extensions.extension import Extension
from office365.entity_collection import EntityCollection
from office365.runtime.resource_path import ResourcePath


class Organization(DirectoryObject):
    """
    The organization resource represents an instance of global settings and resources
    which operate and are provisioned at the tenant-level.
    """

    @property
    def extensions(self):
        """The collection of open extensions defined for the message. Nullable."""
        return self.properties.get('extensions',
                                   EntityCollection(self.context, Extension,
                                                    ResourcePath("extensions", self.resource_path)))



from office365.runtime.client_value import ClientValue


class ListItemCreationInformation(ClientValue):

    def __init__(self, leafName=None, folderUrl=None, underlyingObjectType=None):
        """
        Specifies the properties of the new list item.

        :param int underlyingObjectType: Specifies whether the new list item is a file or a folder.
        :param str leafName: Specifies the name of the new list item. It MUST be the name of the file if the parent
            list of the list item is a document library.
        :param str folderUrl: Specifies the folder for the new list item. It MUST be NULL, empty, a server-relative
            URL, or an absolute URL. If the value is a server-relative URL or an absolute URL, it MUST be under the root
            folder of the list.



        """
        super(ListItemCreationInformation, self).__init__()
        self.FolderUrl = folderUrl
        self.LeafName = leafName
        self.UnderlyingObjectType = underlyingObjectType



from office365.directory.directory_object import DirectoryObject


class OrgContact(DirectoryObject):
    """Represents an organizational contact. Organizational contacts are managed by an organization's administrators
    and are different from personal contacts. Additionally, organizational contacts are either synchronized
    from on-premises directories or from Exchange Online, and are read-only."""
    pass



from office365.directory.identities.identity_set import IdentitySet
from office365.sharepoint.base_entity import BaseEntity


class ChangeTrackedEntity(BaseEntity):

    @property
    def created_datetime(self):
        return self.properties.get('createdDateTime', None)

    @property
    def last_modified_datetime(self):
        return self.properties.get('lastModifiedDateTime', None)

    @property
    def last_modified_by(self):
        return self.properties.get('lastModifiedBy', IdentitySet())
